#
# This file is part of snmpsim software.
#
# Copyright (c) 2010-2019, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/snmpsim/license.html
#
# Managed value variation module
# Send SNMP Notification
#
from pysnmp.hlapi.asyncio import *

from snmpsim import error
from snmpsim import log
from snmpsim.grammar.snmprec import SnmprecGrammar
from snmpsim.record.snmprec import SnmprecRecord
from snmpsim.utils import run_in_new_loop, split


def init(**context):
    pass


TYPE_MAP = {
    "s": OctetString,
    "h": lambda x: OctetString(hexValue=x),
    "i": Integer32,
    "o": ObjectIdentifier,
    "a": IpAddress,
    "u": Unsigned32,
    "g": Gauge32,
    "t": TimeTicks,
    "b": Bits,
    "I": Counter64,
}

MODULE_OPTIONS = (
    ("op", "set"),
    ("community", "public"),
    ("authkey", None),
    ("authproto", "md5"),
    ("privkey", None),
    ("privproto", "des"),
    ("proto", "udp"),
    ("port", "162"),
    ("ntftype", "trap"),
    ("trapoid", "1.3.6.1.6.3.1.1.5.1"),
)


def _cbFun(
    snmpEngine,
    sendRequestHandle,
    errorIndication,
    errorStatus,
    errorIndex,
    varBinds,
    cbCtx,
):
    oid, value = cbCtx

    if errorIndication or errorStatus:
        log.info(
            "notification: for %s=%r failed with errorIndication %s, "
            "errorStatus %s" % (oid, value, errorIndication, errorStatus)
        )


def variate(oid, tag, value, **context):
    if "snmpEngine" in context and context["snmpEngine"]:
        snmpEngine = context["snmpEngine"]

        if snmpEngine not in moduleContext:
            moduleContext[snmpEngine] = {}

        if context["transportDomain"] not in moduleContext[snmpEngine]:
            # register this SNMP Engine to handle our transports'
            # receiver IDs (which we build by outbound and simulator
            # transportDomains concatenation)
            snmpEngine.registerTransportDispatcher(
                snmpEngine.transportDispatcher,
                UdpTransportTarget.transportDomain + context["transportDomain"],
            )

            snmpEngine.registerTransportDispatcher(
                snmpEngine.transportDispatcher,
                Udp6TransportTarget.transportDomain + context["transportDomain"],
            )

            moduleContext[snmpEngine][context["transportDomain"]] = 1

    else:
        raise error.SnmpsimError(
            "Variation module is not given snmpEngine. "
            "Make sure you are not running in --v2c-arch mode"
        )

    if not context["nextFlag"] and not context["exactMatch"]:
        return context["origOid"], tag, context["errorStatus"]

    if "settings" not in recordContext:
        recordContext["settings"] = dict([split(x, "=") for x in split(value, ",")])

        for k, v in MODULE_OPTIONS:
            recordContext["settings"].setdefault(k, v)

        if "hexvalue" in recordContext["settings"]:
            recordContext["settings"]["value"] = [
                int(recordContext["settings"]["hexvalue"][x : x + 2], 16)
                for x in range(0, len(recordContext["settings"]["hexvalue"]), 2)
            ]

        if "vlist" in recordContext["settings"]:
            vlist = {}

            recordContext["settings"]["vlist"] = split(
                recordContext["settings"]["vlist"], ":"
            )

            while recordContext["settings"]["vlist"]:
                o, v = recordContext["settings"]["vlist"][:2]

                vlist = recordContext["settings"]["vlist"][2:]
                recordContext["settings"]["vlist"] = vlist

                typeTag, _ = SnmprecRecord.unpack_tag(tag)

                v = SnmprecGrammar.TAG_MAP[typeTag](v)

                if o not in vlist:
                    vlist[o] = set()

                if o == "eq":
                    vlist[o].add(v)

                elif o in ("lt", "gt"):
                    vlist[o] = v

                else:
                    log.info(
                        "notification: bad vlist syntax: "
                        "%s" % recordContext["settings"]["vlist"]
                    )

            recordContext["settings"]["vlist"] = vlist

    args = recordContext["settings"]

    if context["setFlag"] and "vlist" in args:
        if "eq" in args["vlist"] and context["origValue"] in args["vlist"]["eq"]:
            pass

        elif "lt" in args["vlist"] and context["origValue"] < args["vlist"]["lt"]:
            pass

        elif "gt" in args["vlist"] and context["origValue"] > args["vlist"]["gt"]:
            pass

        else:
            return oid, tag, context["origValue"]

    if args["op"] not in ("get", "set", "any", "*"):
        log.info(
            "notification: unknown SNMP request type configured: " "%s" % args["op"]
        )
        return context["origOid"], tag, context["errorStatus"]

    if (
        args["op"] == "get"
        and not context["setFlag"]
        or args["op"] == "set"
        and context["setFlag"]
        or args["op"] in ("any", "*")
    ):
        if args["version"] in ("1", "2c"):
            authData = CommunityData(
                args["community"], mpModel=args["version"] == "2c" and 1 or 0
            )

        elif args["version"] == "3":
            if args["authproto"] == "md5":
                authProtocol = usmHMACMD5AuthProtocol

            elif args["authproto"] == "sha":
                authProtocol = usmHMACSHAAuthProtocol

            elif args["authproto"] == "none":
                authProtocol = usmNoAuthProtocol

            else:
                log.info("notification: unknown auth proto " "%s" % args["authproto"])
                return context["origOid"], tag, context["errorStatus"]

            if args["privproto"] == "des":
                privProtocol = usmDESPrivProtocol

            elif args["privproto"] == "aes":
                privProtocol = usmAesCfb128Protocol

            elif args["privproto"] == "none":
                privProtocol = usmNoPrivProtocol

            else:
                log.info(
                    "notification: unknown privacy proto " "%s" % args["privproto"]
                )
                return context["origOid"], tag, context["errorStatus"]

            authData = UsmUserData(
                args["user"],
                args["authkey"],
                args["privkey"],
                authProtocol=authProtocol,
                privProtocol=privProtocol,
            )

        else:
            log.info("notification: unknown SNMP version %s" % args["version"])
            return context["origOid"], tag, context["errorStatus"]

        if "host" not in args:
            log.info(
                "notification: target hostname not configured for " "OID %s" % (oid,)
            )
            return context["origOid"], tag, context["errorStatus"]

        if args["proto"] == "udp":
            target = UdpTransportTarget((args["host"], int(args["port"])))

        elif args["proto"] == "udp6":
            target = Udp6TransportTarget((args["host"], int(args["port"])))

        else:
            log.info("notification: unknown transport %s" % args["proto"])
            return context["origOid"], tag, context["errorStatus"]

        localAddress = None

        if "bindaddr" in args:
            localAddress = args["bindaddr"]

        else:
            transportDomain = context["transportDomain"][: len(target.transportDomain)]
            if transportDomain == target.transportDomain:
                # localAddress = snmpEngine.transportDispatcher.getTransport(
                #     context["transportDomain"]
                # ).getLocalAddress()[0]
                pass

            else:
                log.info(
                    "notification: incompatible network transport types used by "
                    "CommandResponder vs NotificationOriginator"
                )

                if "bindaddr" in args:
                    localAddress = args["bindaddr"]

        if localAddress:
            log.info("notification: binding to local address %s" % localAddress)
            target.setLocalAddress((localAddress, 0))

        # this will make target objects different based on their bind address
        target.transportDomain = target.transportDomain + context["transportDomain"]

        varBinds = []

        if "uptime" in args:
            varBinds.append(
                (ObjectIdentifier("1.3.6.1.2.1.1.3.0"), TimeTicks(args["uptime"]))
            )

        if args["version"] == "1":
            if "agentaddress" in args:
                varBinds.append(
                    (
                        ObjectIdentifier("1.3.6.1.6.3.18.1.3.0"),
                        IpAddress(args["agentaddress"]),
                    )
                )

            if "enterprise" in args:
                varBinds.append(
                    (
                        ObjectIdentifier("1.3.6.1.6.3.1.1.4.3.0"),
                        ObjectIdentifier(args["enterprise"]),
                    )
                )

        if "varbinds" in args:
            vbs = split(args["varbinds"], ":")
            while vbs:
                varBinds.append((ObjectIdentifier(vbs[0]), TYPE_MAP[vbs[1]](vbs[2])))
                vbs = vbs[3:]

        notificationType = NotificationType(
            ObjectIdentity(args["trapoid"])
        ).addVarBinds(*varBinds)

        run_in_new_loop(
            sendNotification(
                snmpEngine,
                authData,
                target,
                ContextData(),
                args["ntftype"],
                notificationType,
                cbFun=_cbFun,
                cbCtx=(oid, value),
            )
        )

        log.info(
            "notification: sending Notification to %s with credentials "
            "%s" % (authData, target)
        )

    if context["setFlag"] or "value" not in args:
        return oid, tag, context["origValue"]

    else:
        return oid, tag, args["value"]


def shutdown(**context):
    pass



#
# This file is part of snmpsim software.
#
# Copyright (c) 2010-2019, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/snmpsim/license.html
#
# Managed value variation module: simulate a writable Agent using
# SQL backend for storing Managed Objects
#
# Module initialization parameters are dbtype:<dbms>,dboptions:<options>
#
# Expects to work a table of the following layout:
# CREATE TABLE <tablename> (oid text, tag text, value text, maxaccess text)
#
from snmpsim import error
from snmpsim import log
from snmpsim.grammar.snmprec import SnmprecGrammar
from snmpsim.utils import split

ISOLATION_LEVELS = {
    "0": "READ UNCOMMITTED",
    "1": "READ COMMITTED",
    "2": "REPEATABLE READ",
    "3": "SERIALIZABLE",
}


def init(**context):
    options = {}

    if context["options"]:
        options.update(dict([split(x, ":") for x in split(context["options"], ",")]))

    if "dbtype" not in options:
        raise error.SnmpsimError("database type not specified")

    db = __import__(
        options["dbtype"], globals(), locals(), options["dbtype"].split(".")[:-1]
    )

    if "dboptions" in options:  # legacy
        connectParams = {"database": options["dboptions"]}

    else:
        connectOpts = (
            "host",
            "port",
            "user",
            "passwd",
            "password",
            "db",
            "database",
            "unix_socket",
            "named_pipe",
        )

        connectParams = {k: options[k] for k in options if k in connectOpts}

        for k in "port", "connect_timeout":
            if k in connectParams:
                connectParams[k] = int(connectParams[k])

    if not connectParams:
        raise error.SnmpsimError("database connect parameters not specified")

    moduleContext["dbConn"] = dbConn = db.connect(**connectParams)
    moduleContext["dbTable"] = dbTable = options.get("dbtable", "snmprec")
    moduleContext["isolationLevel"] = options.get("isolationlevel", "1")

    if moduleContext["isolationLevel"] not in ISOLATION_LEVELS:
        raise error.SnmpsimError(
            "unknown SQL transaction isolation level "
            "%s" % moduleContext["isolationLevel"]
        )

    if "mode" in context and context["mode"] == "recording":
        cursor = dbConn.cursor()

        try:
            cursor.execute(
                "select * from information_schema.tables "
                "where table_name='%s' limit 1" % dbTable
            )

        except Exception:  # non-ANSI database
            try:
                cursor.execute("select * from %s limit 1" % dbTable)

            except Exception:
                createTable = True

            else:
                createTable = False

        else:
            createTable = not cursor.fetchone()

        if createTable:
            cursor.execute(
                "CREATE TABLE %s (oid text, tag text, value text, "
                "maxaccess text)" % dbTable
            )

        cursor.close()


def variate(oid, tag, value, **context):
    if "dbConn" in moduleContext:
        db_conn = moduleContext["dbConn"]

    else:
        raise error.SnmpsimError("variation module not initialized")

    cursor = db_conn.cursor()

    try:
        cursor.execute(
            "set session transaction isolation level "
            "%s" % ISOLATION_LEVELS[moduleContext["isolationLevel"]]
        )
        cursor.fetchall()

    except Exception:  # non-MySQL/Postgres
        pass

    if value:
        db_table = value.split(",").pop(0)

    elif "dbTable" in moduleContext:
        db_table = moduleContext["dbTable"]

    else:
        log.info("SQL table not specified for OID " "%s" % (context["origOid"],))
        return context["origOid"], tag, context["errorStatus"]

    orig_oid = context["origOid"]
    sql_oid = ".".join(["%10s" % x for x in str(orig_oid).split(".")])

    if context["setFlag"]:
        if "hexvalue" in context:
            text_tag = context["hextag"]
            text_value = context["hexvalue"]

        else:
            text_tag = SnmprecGrammar().get_tag_by_type(context["origValue"])
            text_value = str(context["origValue"])

        cursor.execute(
            "select maxaccess,tag from %s where oid='%s'"
            " limit 1" % (db_table, sql_oid)
        )

        resultset = cursor.fetchone()

        if resultset:
            maxaccess = resultset[0]
            if maxaccess != "read-write":
                return orig_oid, tag, context["errorStatus"]

            cursor.execute(
                "update %s set tag='%s',value='%s' where "
                "oid='%s'" % (db_table, text_tag, text_value, sql_oid)
            )

        else:
            cursor.execute(
                "insert into %s values ('%s', '%s', '%s', "
                "'read-write')" % (db_table, sql_oid, text_tag, text_value)
            )

        if context["varsRemaining"] == 0:  # last OID in PDU
            db_conn.commit()

        cursor.close()

        return orig_oid, text_tag, context["origValue"]

    else:
        if context["nextFlag"]:
            cursor.execute(
                "select oid from %s where oid>'%s' order by oid "
                "limit 1" % (db_table, sql_oid)
            )

            resultset = cursor.fetchone()

            if resultset:
                orig_oid = orig_oid.clone(
                    ".".join([x.strip() for x in str(resultset[0]).split(".")])
                )

                sql_oid = ".".join(["%10s" % x for x in str(orig_oid).split(".")])

            else:
                cursor.close()
                return orig_oid, tag, context["errorStatus"]

        cursor.execute(
            "select tag, value from %s where oid='%s' " "limit 1" % (db_table, sql_oid)
        )

        resultset = cursor.fetchone()

        cursor.close()

        if resultset:
            return orig_oid, str(resultset[0]), str(resultset[1])

        else:
            return orig_oid, tag, context["errorStatus"]


def record(oid, tag, value, **context):
    if "dbConn" in moduleContext:
        db_conn = moduleContext["dbConn"]

    else:
        raise error.SnmpsimError("variation module not initialized")

    db_table = moduleContext["dbTable"]

    if context["stopFlag"]:
        raise error.NoDataNotification()

    sql_oid = ".".join(["%10s" % x for x in oid.split(".")])
    if "hexvalue" in context:
        text_tag = context["hextag"]
        text_value = context["hexvalue"]

    else:
        text_tag = SnmprecGrammar().get_tag_by_type(context["origValue"])
        text_value = str(context["origValue"])

    cursor = db_conn.cursor()

    cursor.execute("select oid from %s where oid='%s' " "limit 1" % (db_table, sql_oid))

    if cursor.fetchone():
        cursor.execute(
            "update %s set tag='%s',value='%s' where "
            "oid='%s'" % (db_table, text_tag, text_value, sql_oid)
        )

    else:
        cursor.execute(
            "insert into %s values ('%s', '%s', '%s', "
            "'read-write')" % (db_table, sql_oid, text_tag, text_value)
        )

    cursor.close()

    if not context["count"]:
        return str(context["startOID"]), ":sql", db_table

    else:
        raise error.NoDataNotification()


def shutdown(**context):
    db_conn = moduleContext.get("dbConn")
    if db_conn:
        if "mode" in context and context["mode"] == "recording":
            db_conn.commit()

        db_conn.close()



#
# This file is part of snmpsim software.
#
# Copyright (c) 2010-2019, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/snmpsim/license.html
#
import os
import sys
import tempfile

if sys.platform[:3] == "win":
    variation = [
        os.path.join(os.environ["HOMEPATH"], "SNMP Simulator", "Variation"),
        os.path.join(os.environ["APPDATA"], "SNMP Simulator", "Variation"),
        os.path.join(os.environ["PROGRAMFILES"], "SNMP Simulator", "Variation"),
        os.path.join(os.path.split(__file__)[0], "variation"),
    ]
    data = [
        os.path.join(os.environ["HOMEPATH"], "SNMP Simulator", "Data"),
        os.path.join(os.environ["APPDATA"], "SNMP Simulator", "Data"),
        os.path.join(os.environ["PROGRAMFILES"], "SNMP Simulator", "Data"),
        os.path.join(os.path.split(__file__)[0], "data"),
    ]

elif sys.platform == "darwin":
    variation = [
        os.path.join(os.environ["HOME"], ".snmpsim", "variation"),
        os.path.join("/", "usr", "local", "share", "snmpsim", "variation"),
        os.path.join(sys.prefix, "snmpsim", "variation"),
        os.path.join(sys.prefix, "share", "snmpsim", "variation"),
        os.path.join(os.path.split(__file__)[0], "variation"),
    ]
    data = [
        os.path.join(os.environ["HOME"], ".snmpsim", "data"),
        os.path.join("/", "usr", "local", "share", "snmpsim", "data"),
        os.path.join(sys.prefix, "snmpsim", "data"),
        os.path.join(sys.prefix, "share", "snmpsim", "data"),
        os.path.join(os.path.split(__file__)[0], "data"),
    ]

else:
    variation = [
        os.path.join(os.environ["HOME"], ".snmpsim", "variation"),
        os.path.join(sys.prefix, "snmpsim", "variation"),
        os.path.join(sys.prefix, "share", "snmpsim", "variation"),
        os.path.join(os.path.split(__file__)[0], "variation"),
    ]
    data = [
        os.path.join(os.environ["HOME"], ".snmpsim", "data"),
        os.path.join(sys.prefix, "snmpsim", "data"),
        os.path.join(sys.prefix, "share", "snmpsim", "data"),
        os.path.join(os.path.split(__file__)[0], "data"),
    ]

cache = os.path.join(tempfile.gettempdir(), "snmpsim")



#
# This file is part of snmpsim software.
#
# Copyright (c) 2010-2019, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/snmpsim/license.html
#
from pyasn1.error import PyAsn1Error
from pyasn1.type import univ

from snmpsim.error import SnmpsimError
from snmpsim.grammar import dump
from snmpsim.record import abstract


class DumpRecord(abstract.AbstractRecord):
    grammar = dump.DumpGrammar()
    ext = "dump"

    def evaluate_oid(self, oid):
        return univ.ObjectIdentifier(oid)

    def evaluate_value(self, oid, tag, value, **context):
        try:
            value = self.grammar.TAG_MAP[tag](value)

        except Exception as exc:
            raise SnmpsimError(
                "value evaluation error for tag %r, value %r: " "%s" % (tag, value, exc)
            )

        # not all callers supply the context - just ignore it
        try:
            if (
                not context["nextFlag"]
                and not context["exactMatch"]
                or context["setFlag"]
            ):
                return context["origOid"], tag, context["errorStatus"]

        except KeyError:
            pass

        return oid, tag, value

    def evaluate(self, line, **context):
        oid, tag, value = self.grammar.parse(line)
        oid = self.evaluate_oid(oid)

        if context.get("oidOnly"):
            value = None

        else:
            try:
                oid, tag, value = self.evaluate_value(oid, tag, value, **context)

            except PyAsn1Error as exc:
                raise SnmpsimError(
                    "value evaluation for %s = %r failed: " "%s\r\n" % (oid, value, exc)
                )

        return oid, value

    def format_oid(self, oid):
        return univ.ObjectIdentifier(oid).prettyPrint()

    def format_value(self, oid, value, **context):
        return self.format_oid(oid), self.grammar.get_tag_by_type(value), str(value)

    def format(self, oid, value, **context):
        return self.grammar.build(*self.format_value(oid, value, **context))



#
# This file is part of snmpsim software.
#
# Copyright (c) 2010-2019, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/snmpsim/license.html
#
from pysnmp.smi import error

from snmpsim import log
from snmpsim.grammar.snmprec import SnmprecGrammar
from snmpsim.record.snmprec import SnmprecRecord
from snmpsim.utils import split

ERROR_TYPES = {
    "toobig": error.TooBigError,
    "nosuchname": error.NoSuchNameError,
    "badvalue": error.BadValueError,
    "readonly": error.ReadOnlyError,
    "generror": error.GenError,
    "noaccess": error.NoAccessError,
    "wrongtype": error.WrongTypeError,
    "wrongvalue": error.WrongValueError,
    "nocreation": error.NoCreationError,
    "inconsistentvalue": error.InconsistentValueError,
    "resourceunavailable": error.ResourceUnavailableError,
    "commitfailed": error.CommitFailedError,
    "undofailed": error.UndoFailedError,
    "authorizationerror": error.AuthorizationError,
    "notwritable": error.NotWritableError,
    "inconsistentname": error.InconsistentNameError,
    # These only makes sense for SNMP v1
    "nosuchobject": error.NoSuchObjectError,
    "nosuchinstance": error.NoSuchInstanceError,
    "endofmib": error.EndOfMibViewError,
}


def init(**context):
    pass


def variate(oid, tag, value, **context):
    if not context["nextFlag"] and not context["exactMatch"]:
        return context["origOid"], tag, context["errorStatus"]

    if "settings" not in recordContext:
        recordContext["settings"] = dict([split(x, "=") for x in split(value, ",")])

        if "hexvalue" in recordContext["settings"]:
            recordContext["settings"]["value"] = [
                int(recordContext["settings"]["hexvalue"][x : x + 2], 16)
                for x in range(0, len(recordContext["settings"]["hexvalue"]), 2)
            ]

        if "status" in recordContext["settings"]:
            recordContext["settings"]["status"] = recordContext["settings"][
                "status"
            ].lower()

        if "op" not in recordContext["settings"]:
            recordContext["settings"]["op"] = "any"

        if "vlist" in recordContext["settings"]:
            vlist = {}

            recordContext["settings"]["vlist"] = split(
                recordContext["settings"]["vlist"], ":"
            )

            while recordContext["settings"]["vlist"]:
                o, v, e = recordContext["settings"]["vlist"][:3]

                recordContext["settings"]["vlist"] = recordContext["settings"]["vlist"][
                    3:
                ]

                typeTag, _ = SnmprecRecord.unpack_tag(tag)

                v = SnmprecGrammar.TAG_MAP[typeTag](v)

                if o not in vlist:
                    vlist[o] = {}

                if o == "eq":
                    vlist[o][v] = e

                elif o in ("lt", "gt"):
                    vlist[o] = v, e

                else:
                    log.info(
                        "error: bad vlist syntax: %s"
                        % recordContext["settings"]["vlist"]
                    )

            recordContext["settings"]["vlist"] = vlist

    e = None

    if context["setFlag"]:
        if "vlist" in recordContext["settings"]:
            if (
                "eq" in recordContext["settings"]["vlist"]
                and context["origValue"] in recordContext["settings"]["vlist"]["eq"]
            ):
                e = recordContext["settings"]["vlist"]["eq"][context["origValue"]]

            elif (
                "lt" in recordContext["settings"]["vlist"]
                and context["origValue"] < recordContext["settings"]["vlist"]["lt"][0]
            ):
                e = recordContext["settings"]["vlist"]["lt"][1]

            elif (
                "gt" in recordContext["settings"]["vlist"]
                and context["origValue"] > recordContext["settings"]["vlist"]["gt"][0]
            ):
                e = recordContext["settings"]["vlist"]["gt"][1]

        elif recordContext["settings"]["op"] in ("set", "any"):
            if "status" in recordContext["settings"]:
                e = recordContext["settings"]["status"]

    else:
        if recordContext["settings"]["op"] in ("get", "any"):
            if "status" in recordContext["settings"]:
                e = recordContext["settings"]["status"]

    if e and e in ERROR_TYPES:
        log.info(f"error: reporting {e} for {oid}")
        raise ERROR_TYPES[e](
            name=oid, idx=max(0, context["varsTotal"] - context["varsRemaining"] - 1)
        )

    if context["setFlag"]:
        recordContext["settings"]["value"] = context["origValue"]

    return oid, tag, recordContext["settings"].get("value", context["errorStatus"])


def shutdown(**context):
    pass



#
# This file is part of snmpsim software.
#
# Copyright (c) 2010-2019, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/snmpsim/license.html
#
# SNMP Agent Simulator
#
import json
import os
import re
import tempfile
import time
import uuid
from functools import wraps

from pyasn1.type import univ
from pysnmp.carrier.asyncio.dgram import udp
from pysnmp.carrier.asyncio.dgram import udp6
from pysnmp.entity import engine

from snmpsim import error
from snmpsim import log
from snmpsim.reporting.formats import base


def camel2snake(name):
    s1 = re.sub(r"(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub(r"([a-z0-9])([A-Z])", r"\1_\2", s1).lower()


def ensure_base_types(f):
    """Convert decorated function's kwargs to Python types.

    Also turn camel-cased keys into snake case.
    """

    def to_base_types(item):
        if isinstance(item, engine.SnmpEngine):
            item = item.snmpEngineID

        if isinstance(item, (univ.Integer, univ.OctetString, univ.ObjectIdentifier)):
            item = item.prettyPrint()

            if item.startswith("0x"):
                item = item[2:]

            return item

        if isinstance(item, (udp.UdpTransportAddress, udp6.Udp6TransportAddress)):
            return str(item[0])

        return item

    def to_dct(dct):
        items = {}

        for k, v in dct.items():
            k = to_base_types(k)
            k = camel2snake(k)

            if isinstance(v, dict):
                v = to_dct(v)

            else:
                v = to_base_types(v)

            items[k] = v

        return items

    @wraps(f)
    def decorated_function(*args, **kwargs):
        return f(*args, **to_dct(kwargs))

    return decorated_function


class NestingDict(dict):
    """Dict with sub-dict as a defaulted value"""

    def __getitem__(self, item):
        try:
            return dict.__getitem__(self, item)

        except KeyError:
            value = self[item] = type(self)()
            return value


class BaseJsonReporter(base.BaseReporter):
    """Common base for JSON-backed family of reporters."""

    REPORTING_PERIOD = 300
    REPORTING_FORMAT = ""
    REPORTING_VERSION = 1
    PRODUCER_UUID = str(uuid.uuid1())

    def __init__(self, *args):
        if not args:
            raise error.SnmpsimError(
                "Missing %s parameter(s). Expected: "
                "<method>:<reports-dir>[:dumping-period]" % self.__class__.__name__
            )

        self._reports_dir = os.path.join(args[0], self.REPORTING_FORMAT)

        if len(args) > 1:
            try:
                self.REPORTING_PERIOD = int(args[1])

            except Exception as exc:
                raise error.SnmpsimError(
                    "Malformed reports dumping period: %s" % args[1]
                )

        try:
            if not os.path.exists(self._reports_dir):
                os.makedirs(self._reports_dir)

        except OSError as exc:
            raise error.SnmpsimError(
                "Failed to create reports directory %s: "
                "%s" % (self._reports_dir, exc)
            )

        self._metrics = NestingDict()
        self._next_dump = time.time() + self.REPORTING_PERIOD

        log.debug(
            "Initialized %s metrics reporter for instance %s, metrics "
            "directory %s, dumping period is %s seconds"
            % (
                self.__class__.__name__,
                self.PRODUCER_UUID,
                self._reports_dir,
                self.REPORTING_PERIOD,
            )
        )

    def flush(self):
        """Dump accumulated metrics into a JSON file.

        Reset all counters upon success.
        """
        if not self._metrics:
            return

        now = int(time.time())

        if self._next_dump > now:
            return

        self._next_dump = now + self.REPORTING_PERIOD

        self._metrics["format"] = self.REPORTING_FORMAT
        self._metrics["version"] = self.REPORTING_VERSION
        self._metrics["producer"] = self.PRODUCER_UUID

        dump_path = os.path.join(self._reports_dir, "%s.json" % now)

        log.debug("Dumping JSON metrics to %s" % dump_path)

        try:
            json_doc = json.dumps(self._metrics, indent=2)

            with tempfile.NamedTemporaryFile(delete=False) as fl:
                fl.write(json_doc.encode("utf-8"))

            os.rename(fl.name, dump_path)

        except Exception as exc:
            log.error("Failure while dumping metrics into " "%s: %s" % (dump_path, exc))

        self._metrics.clear()


class MinimalJsonReporter(BaseJsonReporter):
    """Collect activity metrics and dump brief report.

    Accumulates and periodically dumps activity metrics reflecting
    SNMP command responder performance.

    These counters are accumulated in memory for some time, then get
    written down as a JSON file indexed by time. Consumers are expected
    to process each of these files and are free to remove them.

    `MinimalJsonReporter` works with both SNMPv1/v2c and SNMPv3
    command responder.

    Activity metrics are arranged as a data structure like this:

    .. code-block:: python

    {
        'format': 'minimaljson',
        'version': 1,
        'producer': <UUID>,
        'first_update': '{timestamp}',
        'last_update': '{timestamp}',
        'transports': {
            'total': 0,
            'failures': 0
        },
        'agents': {
            'total': 0,
            'failures': 0
        },
        'data_files': {
            'total': 0,
            'failures': 0
        }
    }
    """

    REPORTING_FORMAT = "minimaljson"

    def update_metrics(self, **kwargs):
        """Process activity update.

        Update internal counters based on activity update information.

        Parameters in `kwargs` serve two purposes: some are used to
        build activity scopes e.g. {transport_domain}->{snmp_engine},
        however those suffixed `*_count` are used to update corresponding
        activity counters that eventually will make their way to
        consumers.
        """

        root_metrics = self._metrics

        metrics = root_metrics

        now = int(time.time())

        if "first_update" not in metrics:
            metrics["first_update"] = now

        metrics["last_update"] = now

        metrics = root_metrics

        try:
            metrics = metrics["transports"]

            metrics["total"] = metrics.get("total", 0) + kwargs.get(
                "transport_call_count", 0
            )
            metrics["failures"] = metrics.get("failures", 0) + kwargs.get(
                "transport_failure_count", 0
            )

        except KeyError:
            pass

        metrics = root_metrics

        try:
            metrics = metrics["data_files"]

            metrics["total"] = metrics.get("total", 0) + kwargs.get(
                "datafile_call_count", 0
            )
            metrics["failures"] = metrics.get("failures", 0) + kwargs.get(
                "datafile_failure_count", 0
            )

            # TODO: some data is still not coming from snmpsim v2carch core

        except KeyError:
            pass


class FullJsonReporter(BaseJsonReporter):
    """Collect activity metrics and dump detailed report.

    Accumulates and periodically dumps activity counters reflecting
    SNMP command responder performance.

    These counters are accumulated in memory for some time, then get
    written down as a JSON file indexed by time. Consumers are expected
    to process each of these files and are free to remove them.

    `FullJsonReporter` can only work within full SNMPv3 command responder.

    Activity metrics are arranged as a data structure like this:

    .. code-block:: python

    {
        'format': 'fulljson',
        'version': 1,
        'producer': <UUID>,
        'first_update': '{timestamp}',
        'last_update': '{timestamp}',
        '{transport_protocol}': {
            '{transport_endpoint}': {  # local address
                'transport_domain': '{transport_domain}',  # endpoint ID
                '{transport_address}', { # peer address
                    'packets': 0,
                    'parse_failures': 0,  # n/a
                    'auth_failures': 0,  # n/a
                    'context_failures': 0, # n/a
                    '{snmp_engine}': {
                        '{security_model}': {
                            '{security_level}': {
                                '{security_name}': {
                                    '{context_engine_id}': {
                                        '{context_name}': {
                                            '{pdu_type}': {
                                                '{data_file}': {
                                                    'pdus': 0,
                                                    'varbinds': 0,
                                                    'failures': 0,
                                                    '{variation_module}': {
                                                        'calls': 0,
                                                        'failures': 0
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    Where `{token}` is replaced with a concrete value taken from request.
    """

    REPORTING_FORMAT = "fulljson"

    @ensure_base_types
    def update_metrics(self, **kwargs):
        """Process activity update.

        Update internal counters based on activity update information.

        Parameters in `kwargs` serve two purposes: some are used to
        build activity scopes e.g. {transport_domain}->{snmp_engine},
        however those suffixed `*_count` are used to update corresponding
        activity counters that eventually will make their way to
        consumers.
        """

        metrics = self._metrics

        now = int(time.time())

        if "first_update" not in metrics:
            metrics["first_update"] = now

        metrics["last_update"] = now

        try:
            metrics = metrics[kwargs["transport_protocol"]]
            metrics = metrics["%s:%s" % kwargs["transport_endpoint"]]
            metrics["transport_domain"] = kwargs["transport_domain"]
            metrics = metrics[kwargs["transport_address"]]
            metrics["packets"] = metrics.get("packets", 0) + kwargs.get(
                "transport_call_count", 0
            )

            # TODO: collect these counters
            metrics["parse_failures"] = 0
            metrics["auth_failures"] = 0
            metrics["context_failures"] = 0

            metrics = metrics[kwargs["snmp_engine"]]
            metrics = metrics[kwargs["security_model"]]
            metrics = metrics[kwargs["security_level"]]
            metrics = metrics[kwargs["security_name"]]
            metrics = metrics[kwargs["context_engine_id"]]
            metrics = metrics[kwargs["pdu_type"]]
            metrics = metrics[kwargs["data_file"]]

            metrics["pdus"] = metrics.get("pdus", 0) + kwargs.get(
                "datafile_call_count", 0
            )
            metrics["failures"] = metrics.get("failures", 0) + kwargs.get(
                "datafile_failure_count", 0
            )
            metrics["varbinds"] = metrics.get("varbinds", 0) + kwargs.get(
                "varbind_count", 0
            )

            metrics = metrics["variations"]
            metrics = metrics[kwargs["variation"]]
            metrics["calls"] = metrics.get("pdus", 0) + kwargs.get(
                "variation_call_count", 0
            )
            metrics["failures"] = metrics.get("failures", 0) + kwargs.get(
                "variation_failure_count", 0
            )

        except KeyError:
            return



#
# This file is part of snmpsim software.
#
# Copyright (c) 2010-2019, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/snmpsim/license.html
#
# SNMP transport endpoints initialization harness
#
import socket

from pysnmp.carrier.asyncio.dgram import udp
from pysnmp.carrier.asyncio.dgram import udp6

from snmpsim.error import SnmpsimError


class TransportEndpointsBase:
    def __init__(self):
        self.__endpoint = None

    def add(self, addr):
        self.__endpoint = self._addEndpoint(addr)
        return self

    def _addEndpoint(self, addr):
        raise NotImplementedError()

    def __len__(self):
        return len(self.__endpoint)

    def __getitem__(self, i):
        return self.__endpoint[i]


class IPv4TransportEndpoints(TransportEndpointsBase):
    def _addEndpoint(self, addr):
        f = lambda h, p=161: (h, int(p))

        try:
            h, p = f(*addr.split(":"))

        except Exception:
            raise SnmpsimError("improper IPv4/UDP endpoint %s" % addr)

        return udp.UdpTransport().openServerMode((h, p)), addr


class IPv6TransportEndpoints(TransportEndpointsBase):
    def _addEndpoint(self, addr):
        if not udp6:
            raise SnmpsimError("This system does not support UDP/IP6")

        if addr.find("]:") != -1 and addr[0] == "[":
            h, p = addr.split("]:")

            try:
                h, p = h[1:], int(p)

            except Exception:
                raise SnmpsimError("improper IPv6/UDP endpoint %s" % addr)

        elif addr[0] == "[" and addr[-1] == "]":
            h, p = addr[1:-1], 161

        else:
            h, p = addr, 161

        return udp6.Udp6Transport().openServerMode((h, p)), addr


def parse_endpoint(arg, ipv6=False):
    address = arg

    # IPv6 notation
    if ipv6 and address.startswith("["):
        address = address.replace("[", "").replace("]", "")

    try:
        if ":" in address:
            address, port = address.rsplit(":", 1)
            port = int(port)

        else:
            port = 161

    except Exception as exc:
        raise SnmpsimError(f"Malformed network endpoint address {arg}: {exc}")

    try:
        address, port = socket.getaddrinfo(
            address,
            port,
            socket.AF_INET6 if ipv6 else socket.AF_INET,
            socket.SOCK_DGRAM,
            socket.IPPROTO_UDP,
        )[0][4][:2]

    except socket.gaierror as exc:
        raise SnmpsimError(f"Unknown hostname {address}: {exc}")

    return address, port



#!/usr/bin/env python3
#
# SNMP Simulator documentation build configuration file, created by
# sphinx-quickstart on Mon Oct 23 17:36:06 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import os

# import sys
# sys.path.insert(0, os.path.abspath('.'))


# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.viewcode",
    "sphinx_sitemap",
    "sphinx_copybutton",
    "notfound.extension",
]

notfound_urls_prefix = "/snmpsim/"

html_baseurl = "https://docs.lextudio.com/snmpsim/"
sitemap_url_scheme = "{link}"
sitemap_suffix_included = False

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "SNMP Simulator"
copyright = "2010-2019, Ilya Etingof. © Copyright 2022-2024, LeXtudio Inc."
author = "LeXtudio Inc. <support@lextudio.com>"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = "1.1"
# The full version, including alpha/beta/rc tags.
release = "1.1.7"

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = []

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"
pygments_dark_style = "monokai"

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "furo"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {
    "source_repository": "https://github.com/lextudio/snmpsim",
    "source_branch": "main",
    "source_directory": "docs/source/",
    "footer_icons": [
        {
            "name": "GitHub",
            "url": "https://github.com/lextudio/snmpsim",
            "html": "",
            "class": "fa-brands fa-solid fa-github fa-2x",
        },
    ],
}

html_css_files = [
    "https://cdnjs.cloudflare.com/ajax/libs/font-awesome/6.0.0/css/fontawesome.min.css",
    "https://cdnjs.cloudflare.com/ajax/libs/font-awesome/6.0.0/css/solid.min.css",
    "https://cdnjs.cloudflare.com/ajax/libs/font-awesome/6.0.0/css/brands.min.css",
]

html_title = "SNMP Simulator Documentation"

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = "_static/favicon.ico"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# This is required for the alabaster theme
# refs: http://alabaster.readthedocs.io/en/latest/installation.html#sidebars
# html_sidebars = {
#     '**': [
#         'relations.html',  # needs 'show_related': True theme option to display
#         'searchbox.html',
#     ]
# }


# If true, links to the reST sources are added to the pages.
# html_show_sourcelink = False

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
# html_show_sphinx = False

# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "SNMPSimulatordoc"


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (
        master_doc,
        "SNMPSimulator.tex",
        "SNMP Simulator Documentation",
        "Ilya Etingof",
        "manual",
    ),
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "snmpsimulator", "SNMP Simulator Documentation", [author], 1)]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "SNMPSimulator",
        "SNMP Simulator Documentation",
        author,
        "SNMPSimulator",
        "One line description of project.",
        "Miscellaneous",
    ),
]


def setup(app):
    on_rtd = os.environ.get("READTHEDOCS", None) == "True"
    if not on_rtd:
        """Insert Google Analytics tracker
        Based on this Stackoverflow suggestion: https://stackoverflow.com/a/41885884
        """
        app.add_js_file("https://www.googletagmanager.com/gtag/js?id=G-DFLYZZK12P")
        app.add_js_file("google_analytics_tracker.js")



#
# This file is part of snmpsim software.
#
# Copyright (c) 2010-2019, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/snmpsim/license.html
#
# Managed value variation module
# Get/set managed value by invoking an external program
#
import subprocess
import sys

from pysnmp.proto import rfc1902

from snmpsim import log
from snmpsim.utils import split


def init(**context):
    moduleContext["settings"] = {}

    if context["options"]:
        moduleContext["settings"].update(
            dict([split(x, ":") for x in split(context["options"], ",")])
        )

    if "shell" not in moduleContext["settings"]:
        moduleContext["settings"]["shell"] = sys.platform[:3] == "win"

    else:
        moduleContext["settings"]["shell"] = int(moduleContext["settings"]["shell"])


def variate(oid, tag, value, **context):
    # in --v2c-arch some of the items are not defined
    transport_domain = transport_address = security_model = "<undefined>"
    security_name = security_level = context_name = transport_domain

    if "transportDomain" in context:
        transport_domain = rfc1902.ObjectName(context["transportDomain"]).prettyPrint()

    if "transportAddress" in context:
        transport_address = ":".join([str(x) for x in context["transportAddress"]])

    if "securityModel" in context:
        security_model = str(context["securityModel"])

    if "securityName" in context:
        security_name = str(context["securityName"])

    if "securityLevel" in context:
        security_level = str(context["securityLevel"])

    if "contextName" in context:
        context_name = str(context["contextName"])

    args = [
        (
            x.replace("@TRANSPORTDOMAIN@", transport_domain)
            .replace("@TRANSPORTADDRESS@", transport_address)
            .replace("@SECURITYMODEL@", security_model)
            .replace("@SECURITYNAME@", security_name)
            .replace("@SECURITYLEVEL@", security_level)
            .replace("@CONTEXTNAME@", context_name)
            .replace("@DATAFILE@", context["dataFile"])
            .replace("@OID@", str(oid))
            .replace("@TAG@", tag)
            .replace("@ORIGOID@", str(context["origOid"]))
            .replace("@ORIGTAG@", str(sum(x for x in context["origValue"].tagSet[0])))
            .replace("@ORIGVALUE@", str(context["origValue"]))
            .replace("@SETFLAG@", str(int(context["setFlag"])))
            .replace("@NEXTFLAG@", str(int(context["nextFlag"])))
            .replace("@SUBTREEFLAG@", str(int(context["subtreeFlag"])))
        )
        for x in split(value, " ")
    ]

    log.info('subprocess: executing external process "%s"' % " ".join(args))

    try:
        handler = subprocess.check_output

    except AttributeError:
        log.info("subprocess: old Python, expect no output!")

        try:
            handler = subprocess.check_call

        except AttributeError:
            handler = subprocess.call

    try:
        return oid, tag, handler(args, shell=moduleContext["settings"]["shell"])

    except getattr(subprocess, "CalledProcessError", Exception):
        log.info("subprocess: external program execution failed")
        return context["origOid"], tag, context["errorStatus"]


def shutdown(**context):
    pass



#
# This file is part of snmpsim software.
#
# Copyright (c) 2010-2019, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/snmpsim/license.html
#
from snmpsim import error


class AbstractGrammar:
    def parse(self, line):
        raise error.SnmpsimError(
            "Method not implemented at %s" % self.__class__.__name__
        )

    def build(self, oid, tag, val):
        raise error.SnmpsimError(
            "Method not implemented at %s" % self.__class__.__name__
        )

    def get_tag_by_type(self, val):
        raise error.SnmpsimError(
            "Method not implemented at %s" % self.__class__.__name__
        )



#
# This file is part of snmpsim software.
#
# Copyright (c) 2010-2019, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/snmpsim/license.html
#
import re

from pyasn1.codec.ber import encoder
from pyasn1.type import univ
from pysnmp.proto import rfc1902

from snmpsim import error
from snmpsim.grammar import abstract


class WalkGrammar(abstract.AbstractGrammar):
    # case-insensitive keys as snmpwalk output tend to vary
    TAG_MAP = {
        "OID:": rfc1902.ObjectName,
        "NULL:": univ.Null,
        "INTEGER:": rfc1902.Integer,
        "STRING:": rfc1902.OctetString,
        "BITS:": rfc1902.Bits,
        "HEX-STRING:": rfc1902.OctetString,
        "GAUGE32:": rfc1902.Gauge32,
        "COUNTER32:": rfc1902.Counter32,
        "COUNTER64:": rfc1902.Counter64,
        "IPADDRESS:": rfc1902.IpAddress,
        "OPAQUE:": rfc1902.Opaque,
        "UNSIGNED32:": rfc1902.Unsigned32,  # this is not needed
        "TIMETICKS:": rfc1902.TimeTicks,  # this is made up
    }

    @staticmethod
    def _integer_filter(value):
        try:
            int(value)
            return value

        except ValueError:
            # Clean enumeration values
            # .1.3.6.1.2.1.2.2.1.3.1 = INTEGER: ethernetCsmacd(6)
            match = re.match(r".*?\((\-?[0-9]+)\)", value)
            if match:
                return match.group(1)

            # Clean values with UNIT suffix
            # .1.3.6.1.2.1.4.13.0 = INTEGER: 60 seconds
            match = re.match(r"(\-?[0-9]+)\s.+", value)
            if match:
                return match.group(1)

            return value

    # possible DISPLAY-HINTs parsing should occur here
    @staticmethod
    def _string_filter(value):
        if not value:
            return value

        elif value[0] == value[-1] == '"':
            return value[1:-1]

        # .1.3.6.1.2.1.2.2.1.6.201752832 = STRING: 60:9c:9f:ec:a3:38
        elif re.match(r"^[0-9a-fA-F]{1,2}(\:[0-9a-fA-F]{1,2})+$", value):
            return [int(x, 16) for x in value.split(":")]

        else:
            return value

    @staticmethod
    def _opaque_filter(value):
        opaque_tag = ""
        match = re.match(r"^(\w+: +)", value)
        if match:
            opaque_tag = match.group(1).upper()
            value = value[len(opaque_tag) :]

        if opaque_tag.startswith("FLOAT: "):
            # .1.3.6.1.4.1.6574.4.3.1.1.0 = Opaque: Float: 100.000000
            return encoder.encode(univ.Real(float(value[7:])))

        elif opaque_tag.startswith("UINT64: "):
            # .1.3.6.1.4.1.2021.10.1.6.2 = Opaque: UInt64: 18446744073709551614
            # there should be BER encoder, but really I not how do this,
            # I just convert to HEX value instead
            hex = f"{int(value):X}"
            add = len(hex) % 2
            hex = add * "0" + hex
            parts = [hex[i : i + 2] for i in range(0, len(hex), 2)]

            # value len
            hex = f"{len(parts):X}"
            add = len(hex) % 2
            hex = add * "0" + hex
            parts = [hex[i : i + 2] for i in range(0, len(hex), 2)] + parts

            # UInt64 subtag is 9f7b
            value = "9F 7B " + " ".join(parts)

        elif opaque_tag.startswith("INT64: "):
            # .1.3.6.1.4.1.2021.10.1.6.3 = Opaque: Int64: 9223372036854775806
            # .1.3.6.1.4.1.2021.10.1.6.4 = Opaque: Int64: -2
            hex = f"{int(value) & (2**64 - 1):X}"
            add = len(hex) % 2
            hex = add * "0" + hex
            parts = [hex[i : i + 2] for i in range(0, len(hex), 2)]

            # value len
            hex = f"{len(parts):X}"
            add = len(hex) % 2
            hex = add * "0" + hex
            parts = [hex[i : i + 2] for i in range(0, len(hex), 2)] + parts

            # Int64 subtag is 9f7a
            value = "9F 7A " + " ".join(parts)

        return [int(y, 16) for y in value.split(" ")]

    @staticmethod
    def _bits_filter(value):
        # rfc1902.Bits does not really initialize from sequences
        if value == "":
            # .1.3.6.1.2.1.10.7.9.1.1.509 = BITS:
            return value

        # Clean bits values
        # .1.3.6.1.2.1.17.6.1.1.1.0 = BITS: 5B 00 00 00   [[...]1 3 4 6 7
        # .1.3.6.1.2.1.17.6.1.1.1.0 = BITS: 5B 00 00 00   clear(1)
        match = re.match(r"^([0-9a-fA-F]{2}(\s+[0-9a-fA-F]{2})*)", value)
        if match:
            return bytes(int(y, 16) for y in match.group(1).split(" "))

        return bytes(int(y, 16) for y in value.split(" "))

    @staticmethod
    def _hex_string_filter(value):
        # .1.3.6.1.2.1.3.1.1.2.2.1.172.30.1.30 = Hex-STRING: 00 C0 FF 43 CE 45   [...C.E]
        match = re.match(r"^([0-9a-fA-F]{2}(\s+[0-9a-fA-F]{2})*)\s+\[", value)
        if match:
            return [int(y, 16) for y in match.group(1).split(" ")]
        elif " " in value:
            return [int(y, 16) for y in value.split(" ")]
        else:
            # This could be the format returned by some unknown windows snmpwalk tool:
            # .1.3.6.1.2.1.2.2.1.6.1 = HEX-STRING: 00029929AE3C
            return [int(value[i : i + 2], 16) for i in range(0, len(value), 2)]

    @staticmethod
    def _gauge_filter(value):
        try:
            int(value)
            return value

        except ValueError:
            # Clean values with UNIT suffix
            # .1.3.6.1.2.1.4.31.1.1.47.1 = Gauge32: 10000 milli-seconds
            match = re.match(r"(\-?[0-9]+)\s.+", value)
            if match:
                return match.group(1)

            return value

    @staticmethod
    def _net_address_filter(value):
        return ".".join([str(int(y, 16)) for y in value.split(":")])

    @staticmethod
    def _time_ticks_filter(value):
        match = re.match(r".*?\(([0-9]+)\)", value)
        if match:
            return match.group(1)

        return value

    def parse(self, line):
        filters = {
            "OPAQUE:": self._opaque_filter,
            "INTEGER:": self._integer_filter,
            "STRING:": self._string_filter,
            "BITS:": self._bits_filter,
            "HEX-STRING:": self._hex_string_filter,
            "GAUGE32:": self._gauge_filter,
            "NETWORK ADDRESS:": self._net_address_filter,
            "TIMETICKS:": self._time_ticks_filter,
        }

        # drop possible 8-bits
        line = line.decode("ascii", "ignore").encode()

        try:
            oid, value = line.decode("iso-8859-1").strip().split(" = ", 1)

        except Exception:
            raise error.SnmpsimError("broken record <%s>" % line)

        if oid and oid[0] == ".":
            oid = oid[1:]

        if value.startswith("Wrong Type (should be"):
            value = value[value.index(": ") + 2 :]

        if value.startswith("No more variables left in this MIB View"):
            value = "STRING: "

        match = re.match(r"^(\w+(?:[\-\ ]\w+)?:)\ ?(.*)", value)
        if match:
            tag = match.group(1)
            value = match.group(2)

        # this is implicit snmpwalk's fuzziness
        elif value == '""' or value == "STRING:":
            tag = "STRING:"
            value = ""

        elif value == "NULL":
            tag = "NULL:"
            value = ""

        else:
            tag = "TimeTicks:"

        if oid and tag:
            handler = filters.get(tag.upper(), lambda x: x)

            # Need rewrite tag for Network Address after filters
            if tag == "Network Address:":
                tag = "IpAddress:"

            return oid, tag.upper(), handler(value.strip())

        raise error.SnmpsimError("broken record <%s>" % line)



#
# This file is part of snmpsim software.
#
# Copyright (c) 2010-2019, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/snmpsim/license.html
#
from pyasn1.type import univ
from pysnmp.proto import rfc1902

from snmpsim import error
from snmpsim.grammar import abstract


class DumpGrammar(abstract.AbstractGrammar):
    TAG_MAP = {
        "0": rfc1902.Counter32,
        "1": rfc1902.Gauge32,
        "2": rfc1902.Integer32,
        "3": rfc1902.IpAddress,
        "4": univ.Null,
        "5": univ.ObjectIdentifier,
        "6": rfc1902.OctetString,
        "7": rfc1902.TimeTicks,
        "8": rfc1902.Counter32,  # an alias
        "9": rfc1902.Counter64,
    }

    @staticmethod
    def _nullFilter(value):
        return ""  # simply drop whatever value is there when it's a Null

    @staticmethod
    def _unhexFilter(value):
        if value[:5].lower() == "hex: ":
            value = [int(x, 16) for x in value[5:].split(".")]

        elif value[0] == '"' and value[-1] == '"':
            value = value[1:-1]

        return value

    def parse(self, line):
        filters = {"4": self._nullFilter, "6": self._unhexFilter}

        try:
            oid, tag, value = line.decode("iso-8859-1").split("|", 2)

        except Exception as exc:
            raise error.SnmpsimError(f"broken record <{line}>: {exc}")

        else:
            if oid and tag:
                handler = filters.get(tag, lambda x: x)
                return oid, tag, handler(value.strip())

            raise error.SnmpsimError("broken record <%s>" % line)



#
# This file is part of snmpsim software.
#
# Copyright (c) 2010-2019, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/snmpsim/license.html
#
import random
import time

from snmpsim import error
from snmpsim import log
from snmpsim.grammar.snmprec import SnmprecGrammar
from snmpsim.record.snmprec import SnmprecRecord
from snmpsim.utils import split


def init(**context):
    random.seed()


def variate(oid, tag, value, **context):
    if not context["nextFlag"] and not context["exactMatch"]:
        return context["origOid"], tag, context["errorStatus"]

    if "settings" not in recordContext:
        recordContext["settings"] = dict([split(x, "=") for x in split(value, ",")])

        if "hexvalue" in recordContext["settings"]:
            recordContext["settings"]["value"] = [
                int(recordContext["settings"]["hexvalue"][x : x + 2], 16)
                for x in range(0, len(recordContext["settings"]["hexvalue"]), 2)
            ]

        if "wait" in recordContext["settings"]:
            recordContext["settings"]["wait"] = float(recordContext["settings"]["wait"])

        else:
            recordContext["settings"]["wait"] = 500.0

        if "deviation" in recordContext["settings"]:
            recordContext["settings"]["deviation"] = float(
                recordContext["settings"]["deviation"]
            )

        else:
            recordContext["settings"]["deviation"] = 0.0

        if "vlist" in recordContext["settings"]:
            vlist = {}

            recordContext["settings"]["vlist"] = split(
                recordContext["settings"]["vlist"], ":"
            )

            while recordContext["settings"]["vlist"]:
                o, v, d = recordContext["settings"]["vlist"][:3]

                recordContext["settings"]["vlist"] = recordContext["settings"]["vlist"][
                    3:
                ]

                d = int(d)

                type_tag, _ = SnmprecRecord.unpack_tag(tag)

                v = SnmprecGrammar.TAG_MAP[type_tag](v)

                if o not in vlist:
                    vlist[o] = {}

                if o == "eq":
                    vlist[o][v] = d

                elif o in ("lt", "gt"):
                    vlist[o] = v, d

                else:
                    log.info(
                        "delay: bad vlist syntax: "
                        "%s" % recordContext["settings"]["vlist"]
                    )

            recordContext["settings"]["vlist"] = vlist

        if "tlist" in recordContext["settings"]:
            tlist = {}

            recordContext["settings"]["tlist"] = split(
                recordContext["settings"]["tlist"], ":"
            )

            while recordContext["settings"]["tlist"]:
                o, v, d = recordContext["settings"]["tlist"][:3]

                recordContext["settings"]["tlist"] = recordContext["settings"]["tlist"][
                    3:
                ]

                v = int(v)
                d = int(d)

                if o not in tlist:
                    tlist[o] = {}

                if o == "eq":
                    tlist[o][v] = d

                elif o in ("lt", "gt"):
                    tlist[o] = v, d

                else:
                    log.info(
                        "delay: bad tlist syntax: "
                        "%s" % recordContext["settings"]["tlist"]
                    )

            recordContext["settings"]["tlist"] = tlist

    if context["setFlag"] and "vlist" in recordContext["settings"]:
        if (
            "eq" in recordContext["settings"]["vlist"]
            and context["origValue"] in recordContext["settings"]["vlist"]["eq"]
        ):
            delay = recordContext["settings"]["vlist"]["eq"][context["origValue"]]

        elif (
            "lt" in recordContext["settings"]["vlist"]
            and context["origValue"] < recordContext["settings"]["vlist"]["lt"][0]
        ):
            delay = recordContext["settings"]["vlist"]["lt"][1]

        elif (
            "gt" in recordContext["settings"]["vlist"]
            and context["origValue"] > recordContext["settings"]["vlist"]["gt"][0]
        ):
            delay = recordContext["settings"]["vlist"]["gt"][1]

        else:
            delay = recordContext["settings"]["wait"]

    elif "tlist" in recordContext["settings"]:
        now = int(time.time())
        if (
            "eq" in recordContext["settings"]["tlist"]
            and now == recordContext["settings"]["tlist"]["eq"]
        ):
            delay = recordContext["settings"]["tlist"]["eq"][now]

        elif (
            "lt" in recordContext["settings"]["tlist"]
            and now < recordContext["settings"]["tlist"]["lt"][0]
        ):
            delay = recordContext["settings"]["tlist"]["lt"][1]

        elif (
            "gt" in recordContext["settings"]["tlist"]
            and now > recordContext["settings"]["tlist"]["gt"][0]
        ):
            delay = recordContext["settings"]["tlist"]["gt"][1]

        else:
            delay = recordContext["settings"]["wait"]

    else:
        delay = recordContext["settings"]["wait"]

    if recordContext["settings"]["deviation"]:
        delay += random.randrange(
            -recordContext["settings"]["deviation"],
            recordContext["settings"]["deviation"],
        )

    if delay < 0:
        delay = 0

    elif delay > 99999:
        log.info("delay: dropping response for %s" % oid)
        raise error.NoDataNotification()

    log.info("delay: waiting %d milliseconds for %s" % (delay, oid))

    time.sleep(delay / 1000)  # ms

    if context["setFlag"] or "value" not in recordContext["settings"]:
        return oid, tag, context["origValue"]

    else:
        return oid, tag, recordContext["settings"]["value"]


def record(oid, tag, value, **context):
    if context["stopFlag"]:
        raise error.NoDataNotification()

    tag += ":delay"
    if "hexvalue" in context:
        text_value = "hexvalue=" + context["hexvalue"]

    else:
        text_value = "value=" + value

    text_value += ",wait=%d" % int((time.time() - context["reqTime"]) * 1000)  # ms

    if "options" in context:
        text_value += "," + context["options"]

    return oid, tag, text_value


def shutdown(**context):
    pass



#
# This file is part of snmpsim software.
#
# Copyright (c) 2010-2019, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/snmpsim/license.html
#
# Managed value variation module: simulate a writable Agent using
# noSQL backend (Redis) for storing Managed Objects
#
# Module initialization parameters are:
#
# host:<redis-host>,port:<redis-port>,db:<redis-db>
#
# Uses the following data layout:
# Redis LIST type containing sorted OIDs. This is used for answering
# GETNEXT/GETBULK type queries
# Redis HASH type containing OID-value pairs
# For successful operation each managed OID must be present in both
# data structures
#
import random
import time

from pysnmp.smi.error import WrongValueError

from snmpsim import error, log
from snmpsim import utils
from snmpsim.grammar.snmprec import SnmprecGrammar
from snmpsim.record.snmprec import SnmprecRecord

redis = utils.try_load("redis")

random.seed()


def init(**context):
    options = {}

    if context["options"]:
        options.update(
            dict([utils.split(x, ":") for x in utils.split(context["options"], ",")])
        )

    connectOpts = "host", "port", "password", "db", "unix_socket"

    connectParams = {k: options[k] for k in options if k in connectOpts}

    for k in "port", "db":
        if k in connectParams:
            connectParams[k] = int(connectParams[k])

    if not connectParams:
        raise error.SnmpsimError("Redis connect parameters not specified")

    if not redis:
        raise error.SnmpsimError("redis-py Python package must be installed!")

    moduleContext["dbConn"] = redis.StrictRedis(**connectParams)

    if context["mode"] == "recording":
        if "key-spaces-id" in options:
            moduleContext["key-spaces-id"] = int(options["key-spaces-id"])

        else:
            moduleContext["key-spaces-id"] = random.randrange(0, 0xFFFFFFFF)

        log.info("redis: using key-spaces-id %s" % moduleContext["key-spaces-id"])

        if "iterations" in options:
            moduleContext["iterations"] = max(0, int(options["iterations"]) - 1)

        if "period" in options:
            moduleContext["period"] = float(options["period"])

        else:
            moduleContext["period"] = 60.0

        redisScript = options.get("evalsha")
        if redisScript:
            log.info("redis: using server-side script %s" % redisScript)

    elif context["mode"] == "variating":
        moduleContext["booted"] = time.time()

    moduleContext["ready"] = True


unpackTag = SnmprecRecord().unpack_tag


# It turned out, that `py-redis` package emits bytes when running
# on Py3 and `str` on Py2. So let's add simple wrappers to all
# Redis calls that return non-ints to overcome this hassle.


def lindex(dbConn, *args):
    ret = dbConn.lindex(*args)
    if ret is not None:
        ret = ret.decode("iso-8859-1")

    return ret


def get(dbConn, *args):
    ret = dbConn.get(*args)
    if ret is not None:
        ret = ret.decode("iso-8859-1")

    return ret


def evalsha(dbConn, *args):
    ret = dbConn.evalsha(*args)
    if ret is not None:
        ret = ret.decode("iso-8859-1")

    return ret


def variate(oid, tag, value, **context):
    if "dbConn" in moduleContext:
        dbConn = moduleContext["dbConn"]

    else:
        raise error.SnmpsimError("variation module not initialized")

    if "settings" not in recordContext:
        settings = recordContext["settings"] = dict(
            [utils.split(x, "=") for x in utils.split(value, ",")]
        )

        if "key-spaces-id" not in settings:
            log.info("redis:mandatory key-spaces-id option is missing")
            return context["origOid"], tag, context["errorStatus"]

        settings["period"] = float(settings.get("period", 60))

        if "evalsha" in settings:
            if not dbConn.script_exists(settings["evalsha"]):
                log.info(
                    "redis: lua script %s does not exist "
                    "at Redis" % settings["evalsha"]
                )
                return context["origOid"], tag, context["errorStatus"]

        recordContext["ready"] = True

    if "ready" not in recordContext:
        return context["origOid"], tag, context["errorStatus"]

    redisScript = recordContext["settings"].get("evalsha")

    keySpacesId = recordContext["settings"]["key-spaces-id"]

    if recordContext["settings"]["period"] and dbConn.llen(keySpacesId):
        booted = time.time() - moduleContext["booted"]
        keySpaceIdx = int(booted) % dbConn.llen(keySpacesId)

    else:
        keySpaceIdx = 0

    keySpace = lindex(dbConn, keySpacesId, keySpaceIdx)

    if (
        "current-keyspace" not in recordContext
        or recordContext["current-keyspace"] != keySpace
    ):
        log.info(
            "redis: now using keyspace %s (cycling period"
            " %s)" % (keySpace, recordContext["settings"]["period"] or "<disabled>")
        )

        recordContext["current-keyspace"] = keySpace

    if keySpace is None:
        return origOid, tag, context["errorStatus"]

    origOid = context["origOid"]
    dbOid = ".".join(["%10s" % x for x in str(origOid).split(".")])

    if context["setFlag"]:
        if "hexvalue" in context:
            textTag = context["hextag"]
            textValue = context["hexvalue"]

        else:
            textTag = SnmprecGrammar().get_tag_by_type(context["origValue"])
            textValue = str(context["origValue"])

        if redisScript:
            prevTagAndValue = evalsha(dbConn, redisScript, 1, keySpace + "-" + dbOid)

        else:
            prevTagAndValue = get(dbConn, keySpace + "-" + dbOid)

        if prevTagAndValue:
            prevTag, prevValue = prevTagAndValue.split("|")

            if unpackTag(prevTag)[0] != unpackTag(textTag)[0]:
                idx = max(0, context["varsTotal"] - context["varsRemaining"] - 1)
                raise WrongValueError(name=origOid, idx=idx)

        else:
            dbConn.linsert(
                keySpace + "-oids_ordering",
                "after",
                getNextOid(dbConn, keySpace, dbOid),
                dbOid,
            )

        if redisScript:
            evalsha(
                dbConn,
                redisScript,
                1,
                keySpace + "-" + dbOid,
                textTag + "|" + textValue,
            )

        else:
            dbConn.set(keySpace + "-" + dbOid, textTag + "|" + textValue)

        return origOid, textTag, context["origValue"]

    else:
        if context["nextFlag"]:
            textOid = lindex(
                dbConn,
                keySpace + "-oids_ordering",
                getNextOid(dbConn, keySpace, dbOid, index=True),
            )

        else:
            textOid = keySpace + "-" + dbOid

        if redisScript:
            tagAndValue = evalsha(dbConn, redisScript, 1, textOid)

        else:
            tagAndValue = get(dbConn, textOid)

        if not tagAndValue:
            return origOid, tag, context["errorStatus"]

        textOid = ".".join([x.strip() for x in textOid.split("-", 1)[1].split(".")])
        textTag, textValue = tagAndValue.split("|", 1)

        return textOid, textTag, textValue


def getNextOid(dbConn, keySpace, dbOid, index=False):
    listKey = keySpace + "-oids_ordering"
    oidKey = keySpace + "-" + dbOid

    maxlen = listsize = dbConn.llen(listKey)
    minlen = 0

    while maxlen >= minlen and listsize:
        listsize -= 1

        idx = minlen + (maxlen - minlen) // 2

        nextOid = lindex(dbConn, listKey, idx)

        if nextOid < oidKey:
            minlen = idx + 1

        elif nextOid > oidKey:
            maxlen = idx - 1

        else:
            idx += 1
            break

    if not listsize:
        raise error.SnmpsimError("empty/unsorted %s" % listKey)

    return not index and lindex(dbConn, listKey, idx) or idx


def record(oid, tag, value, **context):
    if "ready" not in moduleContext:
        raise error.SnmpsimError("module not initialized")

    if "dbConn" in moduleContext:
        dbConn = moduleContext["dbConn"]

    else:
        raise error.SnmpsimError("variation module not initialized")

    if "started" not in moduleContext:
        moduleContext["started"] = time.time()

    redisScript = moduleContext.get("evalsha")

    keySpace = "%.10d" % (
        moduleContext["key-spaces-id"] + moduleContext.get("iterations", 0)
    )

    if context["stopFlag"]:
        dbConn.sort(
            keySpace + "-" + "temp_oids_ordering",
            store=keySpace + "-" + "oids_ordering",
            alpha=True,
        )

        dbConn.delete(keySpace + "-" + "temp_oids_ordering")
        dbConn.rpush(moduleContext["key-spaces-id"], keySpace)

        log.info("redis: done with key-space %s" % keySpace)

        if "iterations" in moduleContext and moduleContext["iterations"]:
            log.info("redis: %s iterations remaining" % moduleContext["iterations"])

            moduleContext["started"] = time.time()
            moduleContext["iterations"] -= 1

            runtime = time.time() - moduleContext["started"]
            wait = max(0, moduleContext["period"] - runtime)

            raise error.MoreDataNotification(period=wait)

        else:
            raise error.NoDataNotification()

    dbOid = ".".join(["%10s" % x for x in oid.split(".")])

    if "hexvalue" in context:
        textTag = context["hextag"]
        textValue = context["hexvalue"]

    else:
        textTag = SnmprecGrammar().get_tag_by_type(context["origValue"])
        textValue = str(context["origValue"])

    dbConn.lpush(keySpace + "-temp_oids_ordering", keySpace + "-" + dbOid)

    if redisScript:
        evalsha(
            dbConn, redisScript, 1, keySpace + "-" + dbOid, textTag + "|" + textValue
        )

    else:
        dbConn.set(keySpace + "-" + dbOid, textTag + "|" + textValue)

    if not context["count"]:
        settings = {"key-spaces-id": moduleContext["key-spaces-id"]}

        if "period" in moduleContext:
            settings["period"] = "%.2f" % float(moduleContext["period"])

        if "addon" in moduleContext:
            settings.update(dict([utils.split(x, "=") for x in moduleContext["addon"]]))

        value = ",".join([f"{k}={v}" for k, v in settings.items()])

        return str(context["startOID"]), ":redis", value

    else:
        raise error.NoDataNotification()


def shutdown(**context):
    if "dbConn" in moduleContext:
        moduleContext.pop("dbConn")



#
# This file is part of snmpsim software.
#
# Copyright (c) 2010-2019, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/snmpsim/license.html
#
import bz2

from snmpsim import error
from snmpsim.grammar import snmprec
from snmpsim.record import dump


class SnmprecRecord(dump.DumpRecord):
    grammar = snmprec.SnmprecGrammar()
    ext = "snmprec"

    # https://docs.python.org/3/reference/lexical_analysis.html#literals
    ESCAPE_CHARS = {
        92: 92,
        39: 39,
        34: 34,
        97: 7,
        98: 8,
        102: 12,
        110: 10,
        114: 13,
        116: 9,
        118: 11,
    }

    @staticmethod
    def unpack_tag(tag):
        if tag.endswith("x") or tag.endswith("e"):
            return tag[:-1], tag[-1]

        else:
            return tag, None

    def evaluate_raw_string(self, escaped):
        """Evaluates raw Python string like `ast.literal_eval` does"""
        unescaped = []
        hexdigit = None
        escape = False

        for char in escaped:
            number = ord(char)

            if hexdigit is not None:
                if hexdigit:
                    number = (int(hexdigit, 16) << 4) + int(char, 16)
                    hexdigit = None

                else:
                    hexdigit = char
                    continue

            if escape:
                escape = False

                try:
                    number = self.ESCAPE_CHARS[number]

                except KeyError:
                    if number == 120:
                        hexdigit = ""
                        continue

                    raise ValueError("Unknown escape character %c" % char)

            elif number == 92:  # '\'
                escape = True
                continue

            unescaped.append(number)

        return unescaped

    def evaluate_value(self, oid, tag, value, **context):
        tag, encoding_id = self.unpack_tag(tag)

        try:
            if encoding_id == "e":
                value = self.evaluate_raw_string(value)

                return oid, tag, self.grammar.TAG_MAP[tag](value)

            elif encoding_id == "x":
                if isinstance(value, bytes):
                    value = value.decode("iso-8859-1")

                return oid, tag, self.grammar.TAG_MAP[tag](hexValue=value)

            else:
                return oid, tag, self.grammar.TAG_MAP[tag](value)

        except Exception as exc:
            raise error.SnmpsimError(
                "value evaluation error for tag %r, value " "%r: %s" % (tag, value, exc)
            )

    def format_value(self, oid, value, **context):
        if "nohex" in context and context["nohex"]:
            hexvalue = None

        else:
            hexvalue = self.grammar.hexify_value(value)

        text_tag = self.grammar.get_tag_by_type(value)

        if hexvalue:
            text_tag, text_value = text_tag + "x", hexvalue

        else:
            try:
                text_value = repr(value.asOctets())

                if text_value.startswith("b"):
                    text_value = text_value[1:]

                text_value = text_value[1:-1]

                if "\\" in text_value:
                    text_tag += "e"

            except AttributeError:
                text_value = str(value)

        return self.format_oid(oid), text_tag, text_value


class CompressedSnmprecRecord(SnmprecRecord):
    ext = "snmprec.bz2"

    @staticmethod
    def open(path, flags="rb"):
        return bz2.BZ2File(path, flags)



#
# This file is part of snmpsim software.
#
# Copyright (c) 2010-2019, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/snmpsim/license.html
#
# SNMP Simulator MIB to data file converter
#
import argparse
import functools
import os
import random
import sys
import traceback

from pyasn1.error import PyAsn1Error
from pyasn1.type import univ
from pysnmp import debug
from pysnmp.proto import rfc1902
from pysnmp.smi import builder
from pysnmp.smi import compiler
from pysnmp.smi import error
from pysnmp.smi import view
from pysnmp.smi.rfc1902 import ObjectIdentity

from snmpsim import utils
from snmpsim.error import SnmpsimError
from snmpsim.record import dump
from snmpsim.record import mvc
from snmpsim.record import sap
from snmpsim.record import snmprec
from snmpsim.record import walk

RECORD_TYPES = {
    dump.DumpRecord.ext: dump.DumpRecord(),
    mvc.MvcRecord.ext: mvc.MvcRecord(),
    sap.SapRecord.ext: sap.SapRecord(),
    walk.WalkRecord.ext: walk.WalkRecord(),
    snmprec.SnmprecRecord.ext: snmprec.SnmprecRecord(),
    snmprec.CompressedSnmprecRecord.ext: snmprec.CompressedSnmprecRecord(),
}

DESCRIPTION = (
    "Converts MIB modules into SNMP simulation data files. "
    "Chooses random values or asks interactively. "
    "Fills SNMP conceptual tables with consistent indices."
)


def _parse_mib_object(arg, last=False):
    if "::" in arg:
        return ObjectIdentity(*arg.split("::", 1), last=last)

    else:
        return univ.ObjectIdentifier(arg)


def _parse_range(arg):
    try:
        minimum, maximum = (int(x) for x in arg.split(","))

    except Exception as exc:
        raise SnmpsimError(f"Malformed integer range {arg}: {exc}")

    return minimum, maximum


def main():
    parser = argparse.ArgumentParser(description=DESCRIPTION)

    parser.add_argument("-v", "--version", action="version", version=utils.TITLE)

    parser.add_argument(
        "--quiet", action="store_true", help="Do not print out informational messages"
    )

    parser.add_argument(
        "--debug",
        choices=debug.flagMap,
        action="append",
        type=str,
        default=[],
        help="Enable one or more categories of SNMP debugging.",
    )

    parser.add_argument(
        "--row-hint", dest="row_hint", action="store_true", help="Hint for MIBs Type"
    )

    parser.add_argument(
        "--mib-source",
        dest="mib_sources",
        metavar="<URI|PATH>",
        action="append",
        type=str,
        default=["http://mibs.pysnmp.com/asn1/@mib@"],
        help="One or more URIs pointing to a collection of ASN.1 MIB files."
        'Optional "@mib@" token gets replaced with desired MIB module '
        "name during MIB search.",
    )

    parser.add_argument(
        "--mib-module",
        dest="mib_modules",
        action="append",
        type=str,
        required=True,
        help="MIB module to generate simulation data from",
    )

    parser.add_argument(
        "--start-object",
        metavar="<MIB::Object|OID>",
        type=_parse_mib_object,
        help="Drop all simulation data records prior to this OID specified "
        "as MIB object (MIB::Object) or OID (1.3.6.)",
    )

    parser.add_argument(
        "--stop-object",
        metavar="<MIB::Object|OID>",
        type=functools.partial(_parse_mib_object, last=True),
        help="Drop all simulation data records after this OID specified "
        "as MIB object (MIB::Object) or OID (1.3.6.)",
    )

    parser.add_argument(
        "--manual-values",
        action="store_true",
        help="Fill all managed objects values interactively",
    )

    parser.add_argument(
        "--automatic-values",
        type=int,
        default=5000,
        help="Probe for suitable managed object value this many times "
        "prior to failing over to manual value specification",
    )

    parser.add_argument(
        "--table-size",
        type=int,
        default=10,
        help="Generate SNMP conceptual tables with this many rows",
    )

    parser.add_argument(
        "--destination-record-type",
        choices=RECORD_TYPES,
        default="snmprec",
        help="Produce simulation data with record of this type",
    )

    parser.add_argument(
        "--output-file",
        metavar="<FILE>",
        type=str,
        help="SNMP simulation data file to write records to",
    )

    parser.add_argument(
        "--string-pool",
        metavar="<words>",
        action="append",
        help="Words to use for simulated string values",
    )

    parser.add_argument(
        "--string-pool-file",
        metavar="<FILE>",
        type=str,
        help="File containing the words for simulating SNMP string values",
    )

    parser.add_argument(
        "--integer32-range",
        metavar="<min,max>",
        type=_parse_range,
        default=(0, 32),
        help="Range of values used to populate simulated Integer32 values",
    )

    parser.add_argument(
        "--unsigned-range",
        metavar="<min,max>",
        type=_parse_range,
        default=(0, 65535),
        help="Range of values used to populate simulated Unsigned values",
    )

    parser.add_argument(
        "--counter-range",
        metavar="<min,max>",
        type=_parse_range,
        default=(0, 0xFFFFFFFF),
        help="Range of values used to populate simulated Counter values",
    )

    parser.add_argument(
        "--counter64-range",
        metavar="<min,max>",
        type=_parse_range,
        default=(0, 0xFFFFFFFFFFFFFFFF),
        help="Range of values used to populate simulated Counter64 values",
    )

    parser.add_argument(
        "--gauge-range",
        metavar="<min,max>",
        type=_parse_range,
        default=(0, 0xFFFFFFFF),
        help="Range of values used to populate simulated Gauge values",
    )

    parser.add_argument(
        "--timeticks-range",
        metavar="<min,max>",
        type=_parse_range,
        default=(0, 0xFFFFFFFF),
        help="Range of values used to populate simulated Timeticks values",
    )

    args = parser.parse_args()

    if args.debug:
        debug.setLogger(debug.Debug(*args.debug))

    if args.manual_values:
        args.automatic_values = 0

    if args.string_pool_file:
        with open(args.string_pool_file) as fl:
            args.string_pool = fl.read().split()

    elif not args.string_pool:
        args.string_pool = [
            "Jaded",
            "zombies",
            "acted",
            "quaintly",
            "but",
            "kept",
            "driving",
            "their",
            "oxen",
            "forward",
        ]

    if args.output_file:
        ext = os.path.extsep + RECORD_TYPES[args.destination_record_type].ext

        if not args.output_file.endswith(ext):
            args.output_file += ext

        args.output_file = RECORD_TYPES[args.destination_record_type].open(
            args.output_file, "wb"
        )

    else:
        args.output_file = sys.stdout

        if sys.version_info >= (3, 0, 0):
            # binary mode write
            args.output_file = sys.stdout.buffer

        elif sys.platform == "win32":
            import msvcrt

            msvcrt.setmode(sys.stdout.fileno(), os.O_BINARY)

    def get_value(syntax, hint="", automatic_values=args.automatic_values):
        make_guess = args.automatic_values

        val = None

        while True:
            if make_guess:
                if isinstance(syntax, rfc1902.IpAddress):
                    val = ".".join([str(random.randrange(1, 256)) for x in range(4)])

                elif isinstance(syntax, rfc1902.TimeTicks):
                    val = random.randrange(
                        args.timeticks_range[0], args.timeticks_range[1]
                    )

                elif isinstance(syntax, rfc1902.Gauge32):
                    val = random.randrange(args.gauge_range[0], args.gauge_range[1])

                elif isinstance(syntax, rfc1902.Counter32):
                    val = random.randrange(args.counter_range[0], args.counter_range[1])

                elif isinstance(syntax, rfc1902.Integer32):
                    val = random.randrange(
                        args.integer32_range[0], args.integer32_range[1]
                    )

                elif isinstance(syntax, rfc1902.Unsigned32):
                    val = random.randrange(
                        args.unsigned_range[0], args.unsigned_range[1]
                    )

                elif isinstance(syntax, rfc1902.Counter64):
                    val = random.randrange(
                        args.counter64_range[0], args.counter64_range[1]
                    )

                elif isinstance(syntax, univ.OctetString):
                    maxWords = 10
                    val = " ".join(
                        [
                            args.string_pool[random.randrange(0, len(args.string_pool))]
                            for i in range(random.randrange(1, maxWords))
                        ]
                    )

                elif isinstance(syntax, univ.ObjectIdentifier):
                    val = ".".join(
                        ["1", "3", "6", "1", "3"]
                        + [
                            "%d" % random.randrange(0, 255)
                            for x in range(random.randrange(0, 10))
                        ]
                    )

                elif isinstance(syntax, rfc1902.Bits):
                    val = [
                        random.randrange(0, 256) for x in range(random.randrange(0, 9))
                    ]

                else:
                    val = "?"

            # remove value enumeration

            try:
                if syntax.tagSet == rfc1902.Integer32.tagSet:
                    return rfc1902.Integer32(syntax.clone(val))

                if syntax.tagSet == rfc1902.Unsigned32.tagSet:
                    return rfc1902.Unsigned32(syntax.clone(val))

                if syntax.tagSet == rfc1902.Bits.tagSet:
                    return rfc1902.OctetString(syntax.clone(val))

                return syntax.clone(val)

            except PyAsn1Error as exc:
                if make_guess == 1:
                    sys.stderr.write(
                        "*** Inconsistent value: %s\r\n*** See constraints and "
                        "suggest a better one for:\r\n" % exc
                    )

                if make_guess:
                    make_guess -= 1
                    continue

            sys.stderr.write(
                "%s# Value ['%s'] ? "
                % (
                    hint,
                    (val is None and "<none>" or val),
                )
            )
            sys.stderr.flush()

            line = sys.stdin.readline().strip()

            if line:
                if line[:2] == "0x":
                    if line[:4] == "0x0x":
                        line = line[2:]

                    elif isinstance(syntax, univ.OctetString):
                        val = syntax.clone(hexValue=line[2:])

                    else:
                        val = int(line[2:], 16)

                else:
                    val = line

    data_file_handler = snmprec.SnmprecRecord()

    mib_builder = builder.MibBuilder()

    # Load MIB tree foundation classes
    (MibScalar, MibTable, MibTableRow, MibTableColumn) = mib_builder.importSymbols(
        "SNMPv2-SMI", "MibScalar", "MibTable", "MibTableRow", "MibTableColumn"
    )

    mib_view_controller = view.MibViewController(mib_builder)

    compiler.addMibCompiler(mib_builder, sources=args.mib_sources)

    try:
        if isinstance(args.start_object, ObjectIdentity):
            args.start_object.resolveWithMib(mib_view_controller)

        if isinstance(args.stop_object, ObjectIdentity):
            args.stop_object.resolveWithMib(mib_view_controller)

    except error.PySnmpError as exc:
        sys.stderr.write("ERROR: %s\r\n" % exc)
        return 1

    output = []

    # MIBs walk
    for modName in args.mib_modules:
        if not args.quiet:
            sys.stderr.write(
                "# MIB module: %s, from %s till "
                "%s\r\n"
                % (
                    modName,
                    args.start_object or "the beginning",
                    args.stop_object or "the end",
                )
            )

        try:
            oid = ObjectIdentity(modName).resolveWithMib(mib_view_controller)

        except error.PySnmpError as exc:
            sys.stderr.write("ERROR: failed on MIB %s: " "%s\r\n" % (modName, exc))
            return 1

        hint = row_hint = ""
        row_oid = None
        suffix = ()
        this_table_size = 0

        while True:
            try:
                oid, label, _ = mib_view_controller.getNextNodeName(oid)

            except error.NoSuchObjectError:
                break

            if row_oid and not row_oid.isPrefixOf(oid):
                this_table_size += 1

                if args.automatic_values:
                    if this_table_size < args.table_size:
                        oid = tuple(row_oid)
                        if not args.quiet:
                            sys.stderr.write(
                                "# Synthesizing row #%d of table %s\r\n"
                                % (this_table_size, row_oid)
                            )

                    else:
                        if not args.quiet:
                            sys.stderr.write(
                                "# Finished table %s (%d rows)\r\n"
                                % (row_oid, this_table_size)
                            )

                        row_oid = None

                else:
                    while True:
                        sys.stderr.write(
                            "# Synthesize row #%d for table %s (y/n)? "
                            % (this_table_size, row_oid)
                        )
                        sys.stderr.flush()

                        line = sys.stdin.readline().strip()
                        if line:
                            if line[0] in ("y", "Y"):
                                oid = tuple(row_oid)
                                break

                            elif line[0] in ("n", "N"):
                                if not args.quiet:
                                    sys.stderr.write(
                                        "# Finished table %s (%d rows)\r\n"
                                        % (row_oid, this_table_size)
                                    )
                                row_oid = None
                                break

            if args.start_object and oid < args.start_object:
                continue  # skip on premature OID

            if args.stop_object and oid > args.stop_object:
                break  # stop on out of range condition

            mib_name, sym_name, _ = mib_view_controller.getNodeLocation(oid)
            (node,) = mib_builder.importSymbols(mib_name, sym_name)

            if isinstance(node, MibTable):
                hint = f"# Table {mib_name}::{sym_name}\r\n"
                if not args.quiet:
                    sys.stderr.write(
                        "# Starting table %s::%s (%s)\r\n"
                        % (mib_name, sym_name, univ.ObjectIdentifier(oid))
                    )
                continue

            elif isinstance(node, MibTableRow):
                row_indices = {}
                suffix = ()
                row_hint = hint + f"# Row {mib_name}::{sym_name}\r\n"

                for implied_flag, idx_mod_name, idx_sym_name in node.getIndexNames():
                    (idxNode,) = mib_builder.importSymbols(idx_mod_name, idx_sym_name)

                    row_hint += "# Index {}::{} (type {})\r\n".format(
                        idx_mod_name,
                        idx_sym_name,
                        idxNode.syntax.__class__.__name__,
                    )

                    row_indices[idxNode.name] = get_value(
                        idxNode.syntax, not args.quiet and row_hint or ""
                    )

                    suffix = suffix + node.getAsName(
                        row_indices[idxNode.name], implied_flag
                    )

                if not row_indices:
                    if not args.quiet:
                        sys.stderr.write(
                            "# WARNING: %s::%s table has no index!\r\n"
                            % (mib_name, sym_name)
                        )

                if row_oid is None:
                    this_table_size = 0

                row_oid = univ.ObjectIdentifier(oid)
                continue

            elif isinstance(node, MibTableColumn):
                oid = node.name
                if oid in row_indices:
                    val = row_indices[oid]

                else:
                    hint = ""
                    if not args.quiet:
                        hint += row_hint
                        hint += "# Column %s::%s (type" " %s)\r\n" % (
                            mib_name,
                            sym_name,
                            node.syntax.__class__.__name__,
                        )

                    val = get_value(node.syntax, hint)

            elif isinstance(node, MibScalar):
                hint = ""
                if not args.row_hint:
                    hint += "# Scalar %s::%s (type %s)" "\r\n" % (
                        mib_name,
                        sym_name,
                        node.syntax.__class__.__name__,
                    )
                oid = node.name
                suffix = (0,)
                val = get_value(node.syntax, hint)

            else:
                hint = ""
                continue

            output.append((oid + suffix, val))

        output.sort(key=lambda x: univ.ObjectIdentifier(x[0]))

        unique = set()

        for oid, val in output:
            if oid in unique:
                if not args.quiet:
                    sys.stderr.write(
                        "# Dropping duplicate OID %s\r\n"
                        % (univ.ObjectIdentifier(oid),)
                    )
            else:
                try:
                    args.output_file.write(data_file_handler.format(oid, val))

                except SnmpsimError as exc:
                    sys.stderr.write(f"ERROR: {exc}\r\n")

                else:
                    unique.add(oid)

        if not args.quiet:
            sys.stderr.write(f"# End of {modName}, {len(unique)} OID(s) dumped\r\n")

    args.output_file.flush()
    args.output_file.close()

    return 0


if __name__ == "__main__":
    try:
        rc = main()

    except KeyboardInterrupt:
        sys.stderr.write("shutting down process...")
        rc = 0

    except Exception as exc:
        sys.stderr.write("process terminated: %s" % exc)

        for line in traceback.format_exception(*sys.exc_info()):
            sys.stderr.write(line.replace("\n", ";"))
        rc = 1

    sys.exit(rc)



#
# This file is part of snmpsim software.
#
# Copyright (c) 2010-2019, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/snmpsim/license.html
#
import sys

from snmpsim import error

if sys.platform[:3] == "win":

    def daemonize(pidfile):
        raise error.SnmpsimError("Windows is not inhabited with daemons!")

    class PrivilegesOf:
        """Context manager performing nothing on Windows"""

        def __init__(self, *args, **kwargs):
            pass

        def __enter__(self):
            pass

        def __exit__(self, *args):
            pass

else:
    import os
    import pwd
    import grp
    import atexit
    import signal
    import tempfile

    def daemonize(pidfile):
        try:
            pid = os.fork()
            if pid > 0:
                # exit first parent
                os._exit(0)

        except OSError as exc:
            raise error.SnmpsimError("ERROR: fork #1 failed: %s" % exc)

        # decouple from parent environment
        try:
            os.chdir("/")
            os.setsid()

        except OSError:
            pass

        os.umask(0)

        # do second fork
        try:
            pid = os.fork()
            if pid > 0:
                # exit from second parent
                os._exit(0)

        except OSError as exc:
            raise error.SnmpsimError("ERROR: fork #2 failed: %s" % exc)

        def signal_cb(s, f):
            raise KeyboardInterrupt

        for s in signal.SIGTERM, signal.SIGINT, signal.SIGHUP, signal.SIGQUIT:
            signal.signal(s, signal_cb)

        # write pidfile
        def atexit_cb():
            try:
                if pidfile:
                    os.remove(pidfile)

            except OSError:
                pass

        atexit.register(atexit_cb)

        try:
            if pidfile:
                fd, nm = tempfile.mkstemp(dir=os.path.dirname(pidfile))
                os.write(fd, ("%d\n" % os.getpid()).encode("utf-8"))
                os.close(fd)
                os.rename(nm, pidfile)

        except Exception as exc:
            raise error.SnmpsimError(f"Failed to create PID file {pidfile}: {exc}")

        # redirect standard file descriptors
        sys.stdout.flush()
        sys.stderr.flush()
        si = open(os.devnull)
        so = open(os.devnull, "a+")
        se = open(os.devnull, "a+")

        os.dup2(si.fileno(), sys.stdin.fileno())
        os.dup2(so.fileno(), sys.stdout.fileno())
        os.dup2(se.fileno(), sys.stderr.fileno())

    class PrivilegesOf:
        """Context manager executing under reduced privileges"""

        def __init__(self, uname, gname, final=False):
            self._uname = uname
            self._gname = gname
            self._final = final
            self._olduid = self._oldgid = None

        def __enter__(self):
            if os.getenv("SNMPSIM_ALLOW_ROOT") == "true":
                return

            if os.getuid() != 0:
                if self._uname or self._gname:
                    try:
                        pw_name = pwd.getpwnam(self._uname).pw_name
                        gr_name = grp.getgrnam(self._gname).gr_name

                    except Exception as exc:
                        raise error.SnmpsimError(
                            "getpwnam()/getgrnam() failed for %s/%s: "
                            "%s" % (self._uname, self._gname, exc)
                        )

                    if self._uname != pw_name or self._gname != gr_name:
                        raise error.SnmpsimError(
                            "Process is running under different UID/GID"
                        )
                else:
                    return

            else:
                if not self._uname or not self._gname:
                    raise error.SnmpsimError(
                        "Must drop privileges to a non-privileged user&group"
                    )

            try:
                runningUid = pwd.getpwnam(self._uname).pw_uid
                runningGid = grp.getgrnam(self._gname).gr_gid

            except Exception as exc:
                raise error.SnmpsimError(
                    "getpwnam()/getgrnam() failed for %s/%s: "
                    "%s" % (self._uname, self._gname, exc)
                )

            try:
                os.setgroups([])

            except Exception as exc:
                raise error.SnmpsimError("setgroups() failed: %s" % exc)

            try:
                if self._final:
                    os.setgid(runningGid)
                    os.setuid(runningUid)

                else:
                    self._olduid = os.getuid()
                    self._oldgid = os.getgid()

                    os.setegid(runningGid)
                    os.seteuid(runningUid)

            except Exception as exc:
                raise error.SnmpsimError(
                    "%s failed for %s/%s: %s"
                    % (
                        self._final and "setgid()/setuid()" or "setegid()/seteuid()",
                        runningGid,
                        runningUid,
                        exc,
                    )
                )

            os.umask(63)  # 0077

        def __exit__(self, *args):
            if self._olduid is None or self._oldgid is None:
                return

            try:
                os.setegid(self._oldgid)
                os.seteuid(self._olduid)

            except Exception as exc:
                raise error.SnmpsimError(
                    "setegid()/seteuid() failed for %s/%s: %s"
                    % (self._oldgid, self._olduid, exc)
                )



#
# This file is part of snmpsim software.
#
# Copyright (c) 2010-2019, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/snmpsim/license.html
#
# SNMP Simulator data file management tool
#
import argparse
import functools
import os
import sys
import traceback

from pyasn1.type import univ
from pysnmp.error import PySnmpError
from pysnmp.smi import builder
from pysnmp.smi import compiler
from pysnmp.smi import rfc1902
from pysnmp.smi import view

from snmpsim import error
from snmpsim import utils
from snmpsim.record import dump
from snmpsim.record import mvc
from snmpsim.record import sap
from snmpsim.record import snmprec
from snmpsim.record import walk
from snmpsim.record.search.file import get_record


class SnmprecRecordMixIn:
    def evaluateValue(self, oid, tag, value, **context):
        # Variation module reference
        if ":" in tag:
            context["backdoor"]["textTag"] = tag
            return oid, "", value

        else:
            return snmprec.SnmprecRecord.evaluate_value(self, oid, tag, value)

    def formatValue(self, oid, value, **context):
        if "textTag" in context["backdoor"]:
            return self.formatOid(oid), context["backdoor"]["textTag"], value

        else:
            return snmprec.SnmprecRecord.format_value(self, oid, value, **context)


class SnmprecRecord(SnmprecRecordMixIn, snmprec.SnmprecRecord):
    pass


class CompressedSnmprecRecord(SnmprecRecordMixIn, snmprec.CompressedSnmprecRecord):
    pass


# data file types and parsers
RECORD_TYPES = {
    dump.DumpRecord.ext: dump.DumpRecord(),
    mvc.MvcRecord.ext: mvc.MvcRecord(),
    sap.SapRecord.ext: sap.SapRecord(),
    walk.WalkRecord.ext: walk.WalkRecord(),
    SnmprecRecord.ext: SnmprecRecord(),
    CompressedSnmprecRecord.ext: CompressedSnmprecRecord(),
}

DESCRIPTION = (
    "SNMP simulation data management and repair tool. Online "
    "documentation at https://www.pysnmp.com/snmpsim"
)


def _parse_mib_object(arg, last=False):
    if "::" in arg:
        return rfc1902.ObjectIdentity(*arg.split("::", 1), last=last)

    else:
        return univ.ObjectIdentifier(arg)


def main():
    lost_comments = False
    written_count = skipped_count = duplicate_count = 0
    broken_count = variation_count = 0

    parser = argparse.ArgumentParser(description=DESCRIPTION)

    parser.add_argument("-v", "--version", action="version", version=utils.TITLE)

    parser.add_argument(
        "--quiet", action="store_true", help="Do not print out informational messages"
    )

    parser.add_argument(
        "--sort-records",
        action="store_true",
        help="Order simulation data records by OID",
    )

    parser.add_argument(
        "--ignore-broken-records",
        action="store_true",
        help="Drop malformed simulation data records rather than bailing out",
    )

    parser.add_argument(
        "--deduplicate-records",
        action="store_true",
        help="Drop duplicate simulation data records",
    )

    parser.add_argument(
        "--escaped-strings",
        action="store_true",
        help='Produce Python-style escaped strings (e.g. "\x00") in '
        "simulation data values rather than hexify such non-ASCII "
        "values",
    )

    parser.add_argument(
        "--start-object",
        dest="start_oid",
        metavar="<MIB::Object|OID>",
        type=_parse_mib_object,
        help="Drop all simulation data records prior to this OID specified "
        "as MIB object (MIB::Object) or OID (1.3.6.)",
    )

    parser.add_argument(
        "--stop-object",
        dest="stop_oid",
        metavar="<MIB::Object|OID>",
        type=functools.partial(_parse_mib_object, last=True),
        help="Drop all simulation data records after this OID specified "
        "as MIB object (MIB::Object) or OID (1.3.6.)",
    )

    parser.add_argument(
        "--mib-source",
        dest="mib_sources",
        metavar="<URI|PATH>",
        action="append",
        type=str,
        help="One or more URIs pointing to a collection of ASN.1 MIB files."
        'Optional "@mib@" token gets replaced with desired MIB module '
        "name during MIB search.",
    )

    parser.add_argument(
        "--source-record-type",
        choices=RECORD_TYPES,
        default="snmprec",
        help="Treat input as simulation data of this record type",
    )

    parser.add_argument(
        "--destination-record-type",
        choices=RECORD_TYPES,
        default="snmprec",
        help="Produce simulation data with record of this type",
    )

    parser.add_argument(
        "--input-file",
        dest="input_files",
        metavar="<FILE>",
        action="append",
        type=str,
        required=True,
        help="SNMP simulation data file to read records from",
    )

    parser.add_argument(
        "--output-file",
        metavar="<FILE>",
        type=str,
        help="SNMP simulation data file to write records to",
    )

    args = parser.parse_args()

    if not args.mib_sources:
        args.mib_sources = ["http://mibs.pysnmp.com/asn1/@mib@"]

    args.input_files = [
        RECORD_TYPES[args.source_record_type].open(x) for x in args.input_files
    ]

    if args.output_file and args.output_file != "-":
        ext = os.path.extsep + RECORD_TYPES[args.destination_record_type].ext

        if not args.output_file.endswith(ext):
            args.output_file += ext

        args.output_file = RECORD_TYPES[args.destination_record_type].open(
            args.output_file, "wb"
        )

    else:
        args.output_file = sys.stdout

        if sys.version_info >= (3, 0, 0):
            # binary mode write
            args.output_file = sys.stdout.buffer

        elif sys.platform == "win32":
            import msvcrt

            msvcrt.setmode(sys.stdout.fileno(), os.O_BINARY)

    if not args.input_files:
        args.input_files.append(sys.stdin)

    if isinstance(args.start_oid, rfc1902.ObjectIdentity) or isinstance(
        args.stop_oid, rfc1902.ObjectIdentity
    ):
        mib_builder = builder.MibBuilder()

        mib_view_controller = view.MibViewController(mib_builder)

        compiler.addMibCompiler(mib_builder, sources=args.mib_sources)

        try:
            if isinstance(args.start_oid, rfc1902.ObjectIdentity):
                args.start_oid.resolveWithMib(mib_view_controller)

            if isinstance(args.stop_oid, rfc1902.ObjectIdentity):
                args.stop_oid.resolveWithMib(mib_view_controller)

        except PySnmpError as exc:
            sys.stderr.write("ERROR: %s\r\n" % exc)
            return 1

    records_list = []

    for input_file in args.input_files:
        if not args.quiet:
            sys.stderr.write(
                "# Input file #%s, processing records from %s till "
                "%s\r\n"
                % (
                    args.input_files.index(input_file),
                    args.start_oid or "the beginning",
                    args.stop_oid or "the end",
                )
            )

        line_no = 0

        while True:
            line, rec_line_no, _ = get_record(input_file, line_no)

            if not line:
                break

            if rec_line_no != line_no + 1:
                if not args.quiet:
                    sys.stderr.write(
                        "# Losing comment at lines %s..%s (input file #"
                        "%s)\r\n"
                        % (
                            line_no + 1,
                            rec_line_no - 1,
                            args.input_files.index(input_file),
                        )
                    )

                line_no = rec_line_no

                lost_comments += 1

            backdoor = {}

            try:
                oid, value = RECORD_TYPES[args.source_record_type].evaluate(
                    line, backdoor=backdoor
                )

            except error.SnmpsimError as exc:
                if args.ignore_broken_records:
                    if not args.quiet:
                        sys.stderr.write(
                            "# Skipping broken record <%s>: " "%s\r\n" % (line, exc)
                        )
                    broken_count += 1
                    continue

                else:
                    if not args.quiet:
                        sys.stderr.write(
                            "ERROR: broken record <%s>: " "%s\r\n" % (line, exc)
                        )

                    return 1

            if (
                args.start_oid
                and args.start_oid > oid
                or args.stop_oid
                and args.stop_oid < oid
            ):
                skipped_count += 1
                continue

            records_list.append((oid, value, backdoor))

    if args.sort_records:
        records_list.sort(key=lambda x: x[0])

    unique_indices = set()

    for record in records_list:
        if args.deduplicate_records:
            if record[0] in unique_indices:
                if not args.quiet:
                    sys.stderr.write(
                        "# Skipping duplicate record " "<%s>\r\n" % record[0]
                    )

                duplicate_count += 1

                continue

            else:
                unique_indices.add(record[0])

        try:
            args.output_file.write(
                RECORD_TYPES[args.destination_record_type].format(
                    record[0], record[1], backdoor=record[2], nohex=args.escaped_strings
                )
            )

        except Exception as exc:
            sys.stderr.write("ERROR: record not written: %s\r\n" % exc)
            break

        written_count += 1

        if record[2]:
            variation_count += 1

    if not args.quiet:
        sys.stderr.write(
            "# Records: written %s, filtered out %s, de-duplicated %s, ignored "
            "%s, broken %s, variated %s\r\n"
            % (
                written_count,
                skipped_count,
                duplicate_count,
                lost_comments,
                broken_count,
                variation_count,
            )
        )

    args.output_file.flush()
    args.output_file.close()

    return 0


if __name__ == "__main__":
    try:
        rc = main()

    except KeyboardInterrupt:
        sys.stderr.write("shutting down process...")
        rc = 0

    except Exception:
        sys.stderr.write("process terminated: %s" % sys.exc_info()[1])

        for line in traceback.format_exception(*sys.exc_info()):
            sys.stderr.write(line.replace("\n", ";"))
        rc = 1

    sys.exit(rc)



#
# This file is part of snmpsim software.
#
# Copyright (c) 2010-2019, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/snmpsim/license.html
#
import logging
import os
import socket
import sys
import time
from logging import handlers

from snmpsim.error import SnmpsimError

LOG_DEBUG = 0
LOG_INFO = 1
LOG_ERROR = 2


class AbstractLogger:
    def __init__(self, progId, *priv):
        self._logger = logging.getLogger(progId)
        self._logger.setLevel(logging.DEBUG)
        self._progId = progId
        self._ident = 0
        self.init(*priv)

    def __call__(self, s):
        self._logger.debug(" " * self._ident + s)

    def inc_ident(self, amount=2):
        self._ident += amount

    def dec_ident(self, amount=2):
        self._ident -= amount
        if self._ident < 0:
            self._ident = 0

    def init(self, *priv):
        pass


class SyslogLogger(AbstractLogger):
    SYSLOG_SOCKET_PATHS = ("/dev/log", "/var/run/syslog")

    def init(self, *priv):
        if len(priv) < 1:
            raise SnmpsimError(
                "Bad syslog params, need at least facility, also accept "
                "host, port, socktype (tcp|udp)"
            )

        if len(priv) < 2:
            priv = [priv[0], "debug"]

        if len(priv) < 3:
            # search for syslog local socket

            for dev in self.SYSLOG_SOCKET_PATHS:
                if os.path.exists(dev):
                    priv = [priv[0], priv[1], dev]
                    break

            else:
                priv = [priv[0], priv[1], "localhost", 514, "udp"]

        if not priv[2].startswith("/"):
            if len(priv) < 4:
                priv = [priv[0], priv[1], priv[2], 514, "udp"]

            if len(priv) < 5:
                priv = [priv[0], priv[1], priv[2], 514, "udp"]

            priv = [priv[0], priv[1], priv[2], int(priv[3]), priv[4]]

        try:
            handler = handlers.SysLogHandler(
                address=(
                    priv[2].startswith("/") and priv[2] or (priv[2], int(priv[3]))
                ),
                facility=priv[0].lower(),
                socktype=(
                    len(priv) > 4
                    and priv[4] == "tcp"
                    and socket.SOCK_STREAM
                    or socket.SOCK_DGRAM
                ),
            )

        except Exception as exc:
            raise SnmpsimError("Bad syslog option(s): %s" % exc)

        handler.setFormatter(logging.Formatter("%(asctime)s %(name)s: %(message)s"))

        self._logger.addHandler(handler)


class FileLogger(AbstractLogger):
    class TimedRotatingFileHandler(handlers.TimedRotatingFileHandler):
        """Store log creation time in a stand-alone file''s mtime"""

        def __init__(self, *args, **kwargs):
            handlers.TimedRotatingFileHandler.__init__(self, *args, **kwargs)

            self._failure = False

            if not os.path.exists(self._filename):
                return

            timestamp = os.stat(self._filename).st_mtime

            # Use a stand-aside file metadata time instead of the last
            # modification of the log file itself, as the stock
            # implementation does.
            # This is to work-around the increasing rotation intervals
            # on process restart.
            self.rollover_at = self.computeRollover(timestamp)

        @property
        def _filename(self):
            return os.path.join(
                os.path.dirname(self.baseFilename),
                "." + os.path.basename(self.baseFilename) + "-timestamp",
            )

        def doRollover(self):
            try:
                handlers.TimedRotatingFileHandler.doRollover(self)

                # note log file creation time
                if os.path.exists(self._filename):
                    os.unlink(self._filename)

                open(self._filename, "w").close()

                self._failure = False

            except OSError as exc:
                # File rotation seems to fail, postpone the next run
                timestamp = time.time()
                self.rollover_at = self.computeRollover(timestamp)

                if not self._failure:
                    self._failure = True
                    error(
                        "Failed to rotate log/timestamp file "
                        "%s: %s" % (self._filename, exc)
                    )

    def init(self, *priv):
        if not priv:
            raise SnmpsimError("Bad log file params, need filename")

        if sys.platform[:3] == "win":
            # fix possibly corrupted absolute windows path
            if len(priv[0]) == 1 and priv[0].isalpha() and len(priv) > 1:
                priv = [priv[0] + ":" + priv[1]] + list(priv[2:])

        maxsize = 0
        maxage = None

        if len(priv) > 1 and priv[1]:
            try:
                if priv[1][-1] == "k":
                    maxsize = int(priv[1][:-1]) * 1024

                elif priv[1][-1] == "m":
                    maxsize = int(priv[1][:-1]) * 1024 * 1024

                elif priv[1][-1] == "g":
                    maxsize = int(priv[1][:-1]) * 1024 * 1024 * 1024

                elif priv[1][-1] == "S":
                    maxage = ("S", int(priv[1][:-1]))

                elif priv[1][-1] == "M":
                    maxage = ("M", int(priv[1][:-1]))

                elif priv[1][-1] == "H":
                    maxage = ("H", int(priv[1][:-1]))

                elif priv[1][-1] == "D":
                    maxage = ("D", int(priv[1][:-1]))

                else:
                    raise ValueError("Unknown log rotation criterion: %s" % priv[1][-1])

            except ValueError:
                raise SnmpsimError(
                    "Error in timed log rotation specification. Use "
                    "<NNN>k,m,g for size or <NNN>S,M,H,D for time limits"
                )

        try:
            if maxsize:
                handler = handlers.RotatingFileHandler(
                    priv[0], backupCount=30, maxBytes=maxsize
                )

            elif maxage:
                handler = self.TimedRotatingFileHandler(
                    priv[0], backupCount=30, when=maxage[0], interval=maxage[1]
                )

            else:
                handler = handlers.WatchedFileHandler(priv[0])

        except Exception as exc:
            raise SnmpsimError("Failure configure logging: %s" % exc)

        handler.setFormatter(logging.Formatter("%(message)s"))

        self._logger.addHandler(handler)

        self(
            "Log file %s, rotation rules: "
            "%s"
            % (
                priv[0],
                maxsize
                and "> %sKB" % (maxsize / 1024)
                or maxage
                and f"{maxage[1]}{maxage[0]}"
                or "<none>",
            )
        )

    def __call__(self, s):
        now = time.time()
        timestamp = time.strftime("%Y-%m-%dT%H:%M:%S", time.localtime(now))
        AbstractLogger.__call__(
            self, "%s.%02d %s: %s" % (timestamp, now % 1 * 100, self._progId, s)
        )


class StreamLogger(AbstractLogger):
    stream = sys.stderr

    def init(self, *priv):
        try:
            handler = logging.StreamHandler(self.stream)

        except AttributeError as exc:
            raise SnmpsimError("Stream logger failure: %s" % exc)

        handler.setFormatter(logging.Formatter("%(message)s"))

        self._logger.addHandler(handler)


class StdoutLogger(StreamLogger):
    stream = sys.stdout


class StderrLogger(StreamLogger):
    stream = sys.stderr


class NullLogger(AbstractLogger):
    def init(self, *priv):
        handler = logging.NullHandler()
        self._logger.addHandler(handler)

    def __call__(self, s):
        pass


METHODS_MAP = {
    "syslog": SyslogLogger,
    "file": FileLogger,
    "stdout": StdoutLogger,
    "stderr": StderrLogger,
    "null": NullLogger,
}

LEVELS_MAP = {
    "debug": LOG_DEBUG,
    "info": LOG_INFO,
    "error": LOG_ERROR,
}

msg = lambda x: None

log_level = LOG_INFO


def error(message, ctx=""):
    if log_level <= LOG_ERROR:
        msg(f"ERROR {message} {ctx}")


def info(message, ctx=""):
    if log_level <= LOG_INFO:
        msg(f"{message} {ctx}")


def debug(message, ctx=""):
    if log_level <= LOG_DEBUG:
        msg(f"DEBUG {message} {ctx}")


def set_level(level):
    global log_level

    try:
        log_level = LEVELS_MAP[level]

    except KeyError:
        raise SnmpsimError(
            'Unknown log level "%s", known levels are: '
            "%s" % (level, ", ".join(LEVELS_MAP))
        )


def set_logger(progId, *priv, **options):
    global msg

    try:
        if not isinstance(msg, AbstractLogger) or options.get("force"):
            msg = METHODS_MAP[priv[0]](progId, *priv[1:])

    except KeyError:
        raise SnmpsimError(
            'Unknown logging method "%s", known methods are: '
            "%s" % (priv[0], ", ".join(METHODS_MAP))
        )



#
# This file is part of snmpsim software.
#
# Copyright (c) 2010-2019, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/snmpsim/license.html
#

import os
import sys

from snmpsim import confdir
from snmpsim import error
from snmpsim import log
from snmpsim import utils
from snmpsim.record.search.file import get_record

dbm = utils.try_load("anydbm")
if dbm:
    whichdb = utils.try_load("whichdb")

else:
    dbm = utils.try_load("dbm")
    whichdb = dbm


class RecordIndex:
    def __init__(self, text_file, text_parser):
        self._text_file = text_file
        self._text_parser = text_parser

        try:
            self._db_file = text_file[: text_file.rindex(os.path.extsep)]

        except ValueError:
            self._db_file = text_file

        self._db_file += os.path.extsep + "dbm"

        self._db_file = os.path.join(
            confdir.cache,
            os.path.splitdrive(self._db_file)[1].replace(os.path.sep, "_"),
        )

        self._db = self._text = None
        self._db_type = "?"

        self._text_file_time = 0

    def __str__(self):
        return "Data file {}, {}-indexed, {}".format(
            self._text_file,
            self._db_type,
            self._db and "opened" or "closed",
        )

    def is_open(self):
        return self._db is not None

    def get_handles(self):
        if self.is_open():
            if self._text_file_time != os.stat(self._text_file)[8]:
                log.info("Text file %s modified, closing" % self._text_file)
                self.close()

        if not self.is_open():
            self.create()
            self.open()

        return self._text, self._db

    @property
    def _db_files(self):
        return (
            self._db_file + os.path.extsep + "db",
            self._db_file + os.path.extsep + "dat",
            self._db_file,
        )

    def create(self, force_index_build=False, validate_data=False):
        text_file_time = os.stat(self._text_file)[8]

        # gdbm on OS X seems to voluntarily append .db, trying to catch that

        index_needed = force_index_build

        for db_file in self._db_files:
            if os.path.exists(db_file):
                if text_file_time < os.stat(db_file)[8]:
                    if index_needed:
                        log.info("Forced index rebuild %s" % db_file)

                    elif not whichdb.whichdb(self._db_file):
                        index_needed = True
                        log.info(
                            "Unsupported index format, rebuilding " "index %s" % db_file
                        )

                else:
                    index_needed = True
                    log.info("Index %s out of date" % db_file)

                break

        else:
            index_needed = True
            log.info(
                "Index %s does not exist for data file "
                "%s" % (self._db_file, self._text_file)
            )

        if index_needed:
            # these might speed-up indexing
            open_flags = "nfu"

            errors = []

            while open_flags:
                try:
                    db = dbm.open(self._db_file, open_flags)

                except Exception as exc:
                    log.debug(
                        'DBM open with flags "%s" failed on file '
                        "%s: %s" % (open_flags, self._db_file, exc)
                    )
                    errors.append(str(exc))
                    open_flags = open_flags[:-1]
                    continue

                else:
                    break
            else:
                raise error.SnmpsimError(
                    "Failed to create %s for data file "
                    "%s: %s" % (self._db_file, self._text_file, "; ".join(errors))
                )

            try:
                text = self._text_parser.open(self._text_file)

            except Exception as exc:
                raise error.SnmpsimError(
                    f"Failed to open data file {self._db_file}: {exc}"
                )

            log.info(
                "Building index %s for data file %s (open flags "
                '"%s")...' % (self._db_file, self._text_file, open_flags)
            )

            sys.stdout.flush()

            line_no = 0
            offset = 0
            prev_offset = -1

            while True:
                line, line_no, offset = get_record(text, line_no, offset)

                if not line:
                    # reference to last OID in data file
                    db["last"] = "%d,%d,%d" % (offset, 0, prev_offset)
                    break

                try:
                    oid, tag, val = self._text_parser.grammar.parse(line)

                except Exception as exc:
                    db.close()

                    for db_file in self._db_files:
                        try:
                            os.remove(db_file)

                        except OSError:
                            pass

                    raise error.SnmpsimError(
                        "Data error at %s:%d:" " %s" % (self._text_file, line_no, exc)
                    )

                if validate_data:
                    try:
                        self._text_parser.evaluate_oid(oid)

                    except Exception as exc:
                        db.close()

                        for db_file in self._db_files:
                            try:
                                os.remove(db_file)

                            except OSError:
                                pass

                        raise error.SnmpsimError(
                            "OID error at %s:%d: %s" % (self._text_file, line_no, exc)
                        )

                    try:
                        self._text_parser.evaluate_value(
                            oid, tag, val, dataValidation=True
                        )

                    except Exception as exc:
                        log.info(
                            "ERROR at line %s, value %r: " "%s" % (line_no, val, exc)
                        )

                # for lines serving subtrees, type is empty in tag field
                db[oid] = "%d,%d,%d" % (offset, tag[0] == ":", prev_offset)

                if tag[0] == ":":
                    prev_offset = offset

                else:
                    prev_offset = -1  # not a subtree - no back reference

                offset += len(line)

            text.close()
            db.close()

            log.info("...%d entries indexed" % line_no)

        self._text_file_time = os.stat(self._text_file)[8]

        self._db_type = whichdb.whichdb(self._db_file)

        return self

    def lookup(self, oid):
        return self._db[oid]

    def open(self):
        self._text = self._text_parser.open(self._text_file)
        self._db = dbm.open(self._db_file)

    def close(self):
        self._text.close()
        self._db.close()
        self._db = self._text = None



#
# This file is part of snmpsim software.
#
# Copyright (c) 2010-2019, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/snmpsim/license.html
#
# Managed value variation module: simulate a numeric value
#
# Valid values in module options are:
#
#   2  - Integer
#   65 - Counter32
#   66 - Gauge32
#   67 - TimeTicks
#   70 - Counter64
#
import math
import random
import time

from pysnmp.proto import rfc1902

from snmpsim import error
from snmpsim import log
from snmpsim.utils import split

BOOTED = time.time()

INTEGER_TYPES = {
    rfc1902.Counter32.tagSet,
    rfc1902.Counter64.tagSet,
    rfc1902.TimeTicks.tagSet,
    rfc1902.Gauge32.tagSet,
    rfc1902.Integer.tagSet,
}


def init(**context):
    if context["mode"] == "variating":
        random.seed()

    if context["mode"] == "recording":
        moduleContext["settings"] = {}

        if context["options"]:
            for x in split(context["options"], ","):
                for k, v in split(x, ":"):
                    if k == "addon":
                        if k in moduleContext["settings"]:
                            moduleContext["settings"][k].append(v)

                        else:
                            moduleContext["settings"][k] = [v]

                    else:
                        moduleContext["settings"][k] = v

                if "iterations" in moduleContext["settings"]:
                    moduleContext["settings"]["iterations"] = int(
                        moduleContext["settings"]["iterations"]
                    )

                    if moduleContext["settings"]["iterations"]:
                        # no reason for more
                        moduleContext["settings"]["iterations"] = 1

                if "period" in moduleContext["settings"]:
                    moduleContext["settings"]["period"] = float(
                        moduleContext["settings"]["period"]
                    )

                else:
                    moduleContext["settings"]["period"] = 10.0

                if "taglist" not in moduleContext["settings"]:
                    moduleContext["settings"]["taglist"] = "2-65-66-67-70"


def variate(oid, tag, value, **context):
    if not context["nextFlag"] and not context["exactMatch"]:
        return context["origOid"], tag, context["errorStatus"]

    if context["setFlag"]:
        return context["origOid"], tag, context["errorStatus"]

    if "settings" not in recordContext:
        recordContext["settings"] = dict([split(x, "=") for x in split(value, ",")])

        for k in recordContext["settings"]:
            if k != "function":
                recordContext["settings"][k] = float(recordContext["settings"][k])

        if "min" not in recordContext["settings"]:
            recordContext["settings"]["min"] = 0

        if "max" not in recordContext["settings"]:
            if tag == "70":
                recordContext["settings"]["max"] = 0xFFFFFFFFFFFFFFFF

            else:
                recordContext["settings"]["max"] = 0xFFFFFFFF

        if "rate" not in recordContext["settings"]:
            recordContext["settings"]["rate"] = 1

        if "function" in recordContext["settings"]:
            f = split(recordContext["settings"]["function"], "%")
            recordContext["settings"]["function"] = getattr(math, f[0]), f[1:]

        else:
            recordContext["settings"]["function"] = lambda x: x, ()

    vold, told = (
        recordContext["settings"].get("initial", recordContext["settings"]["min"]),
        BOOTED,
    )

    if "cumulative" in recordContext["settings"]:
        if "value" not in recordContext:
            recordContext["value"] = vold, told

        vold, told = recordContext["value"]

    tnow = time.time()

    if "atime" in recordContext["settings"]:
        t = tnow

    else:
        t = tnow - BOOTED

    f, args = recordContext["settings"]["function"]

    _args = []

    if args:
        for x in args:
            if x == "<time>":
                _args.append(t * recordContext["settings"]["rate"])

            else:
                _args.append(float(x))

    else:
        _args.append(t * recordContext["settings"]["rate"])

    v = f(*_args)

    if "scale" in recordContext["settings"]:
        v *= recordContext["settings"]["scale"]

    if "offset" in recordContext["settings"]:
        if "cumulative" in recordContext["settings"]:
            rate = recordContext["settings"]["rate"]
            v += recordContext["settings"]["offset"] * (tnow - told) * rate

        else:
            v += recordContext["settings"]["offset"]

    deviation = recordContext["settings"].get("deviation")
    if deviation:
        v += random.randrange(-deviation, deviation)

    if "cumulative" in recordContext["settings"]:
        v = max(v, 0)

    v += vold

    if v < recordContext["settings"]["min"]:
        v = recordContext["settings"]["min"]

    elif v > recordContext["settings"]["max"]:
        if "wrap" in recordContext["settings"]:
            v %= recordContext["settings"]["max"]
            v += recordContext["settings"]["min"]

        else:
            v = recordContext["settings"]["max"]

    if "cumulative" in recordContext["settings"]:
        recordContext["value"] = v, tnow

    return oid, tag, v


def record(oid, tag, value, **context):
    if "started" not in moduleContext:
        moduleContext["started"] = time.time()

    if "iterations" not in moduleContext:
        moduleContext["iterations"] = min(
            1, moduleContext["settings"].get("iterations", 0)
        )

    # single-run recording

    iterations = moduleContext["settings"].get("iterations")
    if not iterations:
        if context["origValue"].tagSet not in INTEGER_TYPES:
            if "hextag" in context:
                tag = context["hextag"]

            if "hexvalue" in context:
                value = context["hexvalue"]

            return oid, tag, value

        if (
            "taglist" not in moduleContext["settings"]
            or tag not in moduleContext["settings"]["taglist"]
        ):
            return oid, tag, value

        value = "initial=%s" % value

        if context["origValue"].tagSet == rfc1902.TimeTicks.tagSet:
            value += ",rate=100"

        elif context["origValue"].tagSet == rfc1902.Integer.tagSet:
            value += ",rate=0"

        return oid, tag + ":numeric", value

    # multiple-iteration recording

    if oid not in moduleContext:
        settings = {"initial": value}

        if context["origValue"].tagSet == rfc1902.TimeTicks.tagSet:
            settings["rate"] = 100

        elif context["origValue"].tagSet == rfc1902.Integer.tagSet:
            settings["rate"] = 0  # may be constants

        if "addon" in moduleContext["settings"]:
            settings.update(
                dict([split(x, "=") for x in moduleContext["settings"]["addon"]])
            )

        moduleContext[oid] = {}

        moduleContext[oid]["settings"] = settings

    if moduleContext["iterations"]:
        if context["stopFlag"]:  # switching to final iteration
            log.info(
                "numeric: %s iterations " "remaining" % moduleContext["iterations"]
            )

            moduleContext["iterations"] -= 1
            moduleContext["started"] = time.time()

            running = time.time() - moduleContext["started"]
            wait = max(0, float(moduleContext["settings"]["period"]) - running)

            raise error.MoreDataNotification(period=wait)

        else:  # storing values on first iteration
            moduleContext[oid]["time"] = time.time()
            moduleContext[oid]["value"] = context["origValue"]

            if "hexvalue" in moduleContext[oid]:
                moduleContext[oid]["hexvalue"] = context["hexvalue"]

            if "hextag" in moduleContext[oid]:
                moduleContext[oid]["hextag"] = context["hextag"]

            raise error.NoDataNotification()
    else:
        if context["stopFlag"]:
            raise error.NoDataNotification()

        if "value" in moduleContext[oid]:
            if context["origValue"].tagSet not in INTEGER_TYPES:
                if "hextag" in moduleContext[oid]:
                    tag = moduleContext[oid]["hextag"]

                if "hexvalue" in moduleContext[oid]:
                    value = moduleContext[oid]["hexvalue"]

                return oid, tag, value

            if tag not in moduleContext["settings"]["taglist"]:
                return oid, tag, moduleContext[oid]["value"]

            diff = int(context["origValue"]) - int(moduleContext[oid]["value"])
            runtime = time.time() - moduleContext[oid]["time"]
            moduleContext[oid]["settings"]["rate"] = diff / runtime

            tag += ":numeric"
            value = ",".join(
                [f"{k}={v}" for k, v in moduleContext[oid]["settings"].items()]
            )

            return oid, tag, value

        else:
            raise error.NoDataNotification()


def shutdown(**context):
    pass



#
# This file is part of snmpsim software.
#
# Copyright (c) 2010-2019, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/snmpsim/license.html
#
# SNMP Agent Simulator
#
from snmpsim import error
from snmpsim.reporting.formats import alljson
from snmpsim.reporting.formats import null
from snmpsim import log


class ReportingManager:
    """Maintain activity metrics.

    Accumulates and periodically dumps activity metrics reflecting
    SNMP command responder activity.

    These counters are accumulated in memory for some time, then get
    written down as a JSON file indexed by time. Consumers are expected
    to process each of these files and are free to remove them.
    """

    REPORTERS = {
        "null": null.NullReporter,
        "fulljson": alljson.FullJsonReporter,
        "minimaljson": alljson.MinimalJsonReporter,
    }

    _reporter = null.NullReporter()

    @classmethod
    def configure(cls, fmt, *args):
        try:
            reporter = cls.REPORTERS[fmt]

        except KeyError:
            raise error.SnmpsimError("Unsupported reporting format: %s" % fmt)

        cls._reporter = reporter(*args)

        log.info(
            'Using "%s" activity reporting method with '
            "params %s" % (cls._reporter, ", ".join(args))
        )

    @classmethod
    def update_metrics(cls, **kwargs):
        cls._reporter.update_metrics(**kwargs)
        cls._reporter.flush()
